"""
hashing algorithms.

Descriptions taken from:
https://raw.githubusercontent.com/micropython/micropython/master/docs/library/hashlib.rst.
====================================

.. module:: hashlib
   :synopsis: hashing algorithms

|see_cpython_module| :mod:`python:hashlib`.

This module implements binary data hashing algorithms. The exact inventory
of available algorithms depends on a board. Among the algorithms which may
be implemented:

* SHA256 - The current generation, modern hashing algorithm (of SHA2 series).
  It is suitable for cryptographically-secure purposes. Included in the
  MicroPython core and any board is recommended to provide this, unless
  it has particular code size constraints.

* SHA1 - A previous generation algorithm. Not recommended for new usages,
  but SHA1 is a part of number of Internet standards and existing
  applications, so boards targeting network connectivity and
  interoperability will try to provide this.

* MD5 - A legacy algorithm, not considered cryptographically secure. Only
  selected boards, targeting interoperability with legacy applications,
  will offer this.
"""

__author__ = "Howard C Lovatt"
__copyright__ = "Howard C Lovatt, 2020 onwards."
__license__ = "MIT https://opensource.org/licenses/MIT (as used by MicroPython)."
__version__ = "7.3.9"  # Version set by https://github.com/hlovatt/tag2ver

from abc import ABC
from typing import overload

from uio import AnyReadableBuf

# noinspection PyPep8Naming
class sha256("_Hash"):
    """
   The current generation, modern hashing algorithm (of SHA2 series).
   It is suitable for cryptographically-secure purposes. Included in the
   MicroPython core and any board is recommended to provide this, unless
   it has particular code size constraints.
   """

    @overload
    def __init__(self):
        """
       Create an SHA256 hasher object and optionally feed ``data`` into it.
      """
    @overload
    def __init__(self, data: AnyReadableBuf):
        """
       Create an SHA256 hasher object and optionally feed ``data`` into it.
      """

# noinspection PyPep8Naming
class sha1("_Hash"):
    """
   A previous generation algorithm. Not recommended for new usages,
   but SHA1 is a part of number of Internet standards and existing
   applications, so boards targeting network connectivity and
   interoperability will try to provide this.
   """

    @overload
    def __init__(self):
        """
       Create an SHA1 hasher object and optionally feed ``data`` into it.
      """
    @overload
    def __init__(self, data: AnyReadableBuf):
        """
       Create an SHA1 hasher object and optionally feed ``data`` into it.
      """

# noinspection PyPep8Naming
class md5("_Hash"):
    """
   A legacy algorithm, not considered cryptographically secure. Only
   selected boards, targeting interoperability with legacy applications,
   will offer this.
   """

    def __init__(self, data: AnyReadableBuf = ..., /):
        """
       Create an MD5 hasher object and optionally feed ``data`` into it.
      """

class _Hash(ABC):
    """
   Abstract base class for hashing algorithms that defines methods available in all algorithms.
   """

    def update(self, data: AnyReadableBuf, /) -> None:
        """
      Feed more binary data into hash.
      """
    def digest(self) -> bytes:
        """
      Return hash for all data passed through hash, as a bytes object. After this
      method is called, more data cannot be fed into the hash any longer.
      """
    def hexdigest(self) -> str:
        """
      This method is NOT implemented. Use ``binascii.hexlify(hash.digest())``
      to achieve a similar effect.
      """



from pybricks.__stub.__control import Control
from pybricks.ev3devices import Motor
from typing import Tuple


class DriveBase:
    """
    A robotic vehicle with two powered wheels and an optional support wheel or caster.

    By specifying the dimensions of your robot, this class makes it easy to drive a given distance in millimeters or turn by a given number of degrees.

    Positive distances and drive speeds mean driving forward. Negative means backward.

    Positive angles and turn rates mean turning right. Negative means left. So when viewed from the top, positive means clockwise and negative means counterclockwise.

    Attributes:
        distance_control (Control): The traveled distance and drive speed are controlled by a PID controller. You can use this attribute to change its settings.
        heading_control (Control): The robot turn angle and turn rate are controlled by a PID controller. You can use this attribute to change its settings.

    Args:
        left_motor (Motor): The motor that drives the left wheel.
        right_motor (Motor): The motor that drives the right wheel.
        wheel_diameter (int): Diameter of the wheels in millimeters.
        axle_track (int): Distance between the points where both wheels touch the ground in millimeters.
    """

    def __init__(self, left_motor: Motor, right_motor: Motor, wheel_diameter: int, axle_track: int):
        self.distance_control = Control()  # type: Control
        self.heading_control = Control()  # type: Control

    def straight(self, distance: int):
        """
        Drives straight for a given distance then stops.

        Args:
            distance (int): Distance to travel in millimeters.
        """
        ...

    def turn(self, angle: int):
        """
        Turns in place by a given angle then stops.

        Args:
            angle (int): Angle of the turn in degrees.
        """
        ...

    def settings(self, straight_speed: int = None, straight_acceleration: int = None, turn_rate: int = None, turn_acceleration: int = None) -> Tuple[int, int, int, int]:
        """
        Configures the speed and acceleration used by straight() and turn().

        If you give no arguments, this returns the current values as a tuple.

        You can only change the settings while the robot is stopped. This is either before you begin driving or after you call stop().

        Args:
            straight_speed (int):  Speed of the robot during straight() in millimeters/second.
            straight_acceleration (int): Acceleration and deceleration of the robot at the start and end of straight() in millimeters/second^2.
            turn_rate (int): Turn rate of the robot during turn() in degrees/second.
            turn_acceleration (int): Angular acceleration and deceleration of the robot at the start and end of turn() in degrees/second^2.

        Returns:
            Straight speed (millimeters/second), straight acceleration (millimeters/second^2), turn rate (degrees/second), and turn acceleration (degrees/second^2) (if no arguments are provided), None otherwise.
        """
        if straight_speed is None and straight_acceleration is None and turn_rate is None and turn_acceleration is None:
            return (0, 0, 0, 0)
        else:
            return None

    def drive(self, drive_speed: int, turn_rate: int):
        """
        Start driving at the specified speed and turnrate. Both values are measured at the center point between the wheels of the robot.

        Args:
            drive_speed (int): Speed of the robot in millimeters/second.
            turn_rate (int): Turn rate of the robot in degrees/second.
        """
        ...

    def stop(self):
        """
        Stops the robot by letting the motors spin freely.
        """
        ...

    def distance(self) -> int:
        """
        Get the estimated driven distance.

        Returns:
            Driven distance since last reset in millimeters.
        """
        return 0

    def angle(self) -> int:
        """
        Get the estimated rotation angle of the drive base.

        Returns:
            Accumulated angle since last reset in degrees.
        """
        return 0

    def state(self) -> Tuple[int, int, int, int]:
        """
        Gets the state of the robot.

        This returns the current distance(), the drive speed, the angle(), and the turn rate.

        Returns:
            Distance in millimeters, Drive Speed in millimeters/second, Angle in degrees, Rotational Speed in degrees/second
        """
        return (0, 0, 0, 0)

    def reset(self):
        """
        Resets the estimated driven distance and angle to 0.
        """
        ...



from tkinter import messagebox

from turcar import get_workbench, tktextext, ui_utils
from turcar.ui_utils import scrollbar_style
import tkinter as tk
from pdf2image import convert_from_path
from PIL import ImageTk, Image

class PdfView(tktextext.TextFrame):
    def __init__(self, master):
        tktextext.TextFrame.__init__(
            self,
            master,
            vertical_scrollbar_style=scrollbar_style("Vertical"),
            horizontal_scrollbar_style=scrollbar_style("Horizontal"),
            horizontal_scrollbar_class=ui_utils.AutoScrollbar,
            read_only=True,
            font="TkDefaultFont",
            padx=10,
            pady=0,
            insertwidth=0,
        )
        self.title = 'pdf预览'
        self.pages = None
        self.current_page_number = 0
        self.current_page_image = None
        self.image_label = None
        self.prev_page_button = None
        self.next_page_button = None
        self.pdf_name = None
        get_workbench().bind("PdfView", self.show_browser)
        get_workbench().bind("HideView", self.hidden_browser)

    def hidden_browser(self, event):
        # self.unbind("PdfView")
        self.unbind("HideView")
        get_workbench().get_view("PdfView").destroy()
        get_workbench().add_view(PdfView, self.title, "se", visible_by_default=False)
    def re_size(self):
        screen_height = self.winfo_screenheight()
        image_height = min(800, screen_height * 0.8)  # 占用80%的屏幕高度
        p_h = screen_height * 0.8
        w = self.pages[self.current_page_number].size[0]
        h = self.pages[self.current_page_number].size[1]
        resized_image = self.pages[self.current_page_number].resize((int(w / (h / image_height)), int(p_h)),
                                                                    Image.LANCZOS)
        return resized_image
    def show_browser(self, event):
        path = event.content
        self.pages = convert_from_path(path)
        self.current_page_number = 0
        self.current_page_image = ImageTk.PhotoImage(self.re_size())
        self.image_label = tk.Label(self.text, image=self.current_page_image)
        self.image_label.grid(row=1, column=0, columnspan=3)

        # 创建“上一个”和“下一个”按钮
        self.prev_page_button = tk.Button(self.text, text="上一页", command=self.prev_page)
        self.prev_page_button.grid(row=0, column=0)

        self.next_page_button = tk.Button(self.text, text="下一页", command=self.next_page)
        self.next_page_button.grid(row=0, column=1)
    def prev_page(self):
        # 显示上一页PDF图像
        if self.current_page_number > 0:
            self.current_page_number -= 1
            self.current_page_image = ImageTk.PhotoImage(self.re_size())
            self.image_label.config(image=self.current_page_image)
        else:
            messagebox.showinfo('第一页', '已经是第一页了哦')

    def next_page(self):
        # 显示下一页PDF图像
        if self.current_page_number < len(self.pages) - 1:
            self.current_page_number += 1
            self.current_page_image = ImageTk.PhotoImage(self.re_size())
            self.image_label.config(image=self.current_page_image)
        else:
            messagebox.showinfo('最后一页', '已经是最后一页了哦')
def init():
    get_workbench().add_view(PdfView, 'pdf预览', "se", visible_by_default=False)


class Tabs(tk.Frame):

    def __init__(self):
        tk.Frame.__init__(self)
        # TODO: implement tabs



from __future__ import annotations

import digitalio
from circuitpython_typing import WriteableBuffer

def flux_readinto(
    buffer: WriteableBuffer, data: digitalio.DigitalInOut, index: digitalio.DigitalInOut
) -> int:
    """Read flux transition information into the buffer.

    The function returns when the buffer has filled, or when the index input
    indicates that one full revolution of data has been recorded.  Due to
    technical limitations, this process may not be interruptible by
    KeyboardInterrupt.

    :param buffer: Read data into this buffer.  Each element represents the time between successive zero-to-one transitions.
    :param data: Pin on which the flux data appears
    :param index: Pin on which the index pulse appears
    :return: The actual number of bytes of read
    """
    ...

def mfm_readinto(
    buffer: WriteableBuffer, data: digitalio.DigitalInOut, index: digitalio.DigitalInOut
) -> int:
    """Read mfm blocks into the buffer.

    The track is assumed to consist of 512-byte sectors.

    The function returns when all sectors have been successfully read, or
    a number of index pulses have occurred.  Due to technical limitations, this
    process may not be interruptible by KeyboardInterrupt.

    :param buffer: Read data into this buffer.  Must be a multiple of 512.
    :param data: Pin on which the mfm data appears
    :param index: Pin on which the index pulse appears
    :return: The actual number of sectors read
    """
    ...

samplerate: int
"""The approximate sample rate in Hz used by flux_readinto."""



from pybricks.parameters import Port
from typing import Callable


class VernierAdapter:
    """
    LEGO® MINDSTORMS® Education NXT/EV3 Adapter for Vernier Sensors.

    Args:
        port (Port): Port to which the sensor is connected.
        conversion (callable): Function of the format conversion(). This function is used to convert the raw analog voltage to the sensor-specific output value. Each Vernier Sensor has its own conversion function. The example given below demonstrates the conversion for the Surface Temperature Sensor.
    """

    def __init__(self, port: Port, conversion: Callable = None):
        if port == Port.A or port == Port.B or port == Port.C or port == Port.D:
            raise ValueError("Sensors must use Port S1, S2, S3, or S4.")

    def voltage(self) -> int:
        """
        Measures the raw analog sensor voltage.

        Returns:
            Analog voltage in millivolts.
        """
        return 0

    def conversion(self, voltage: int) -> int:
        """
        Converts the raw voltage (mV) to a sensor value.

        If you did not provide a conversion function earlier, no conversion will be applied.

        Args:
            voltage (int): Analog sensor voltage in millivolts.

        Returns:
            Converted sensor value.
        """
        return 0

    def value(self) -> int:
        """
        Measures the sensor voltage() and then applies your conversion() to give you the sensor value.

        Returns:
            Converted sensor value.
        """
        return 0



from typing import List

def listdir() -> List[str]:
    """
    Returns a list of the names of all the files contained within the local persistent on-device file system.
    """


def remove(filename: str) -> None:
    """
    Removes (deletes) the file named in the argument filename. If the file does not exist an OSError exception will occur.
    """


def size(filename: str) -> int:
    """
    Returns the size, in bytes, of the file named in the argument filename. If the file does not exist an OSError exception will occur.
    """


def uname():
    """
    Returns information identifying the current operating system. The return value is an object with five attributes:

    * sysname - operating system name
    * nodename - name of machine on network (implementation-defined)
    * release - operating system release
    * version - operating system version
    * machine - hardware identifier
    """



"""
Low-level Bluetooth radio functionality.

Descriptions taken from:
https://raw.githubusercontent.com/micropython/micropython/master/docs/library/bluetooth.rst.

This module provides an interface to a Bluetooth controller on a board.
Currently this supports Bluetooth Low Energy (BLE) in Central, Peripheral,
Broadcaster, and Observer roles, as well as GATT Server and Client and L2CAP
connection-oriented-channels. A device may operate in multiple roles
concurrently. Pairing (and bonding) is supported on some ports.

This API is intended to match the low-level Bluetooth protocol and provide
building-blocks for higher-level abstractions such as specific device types.

.. note:: This module is still under development and its classes, functions,
          methods and constants are subject to change.
"""

__author__ = "Howard C Lovatt"
__copyright__ = "Howard C Lovatt, 2020 onwards."
__license__ = "MIT https://opensource.org/licenses/MIT (as used by MicroPython)."
__version__ = "7.3.9"  # Version set by https://github.com/hlovatt/tag2ver

from typing import overload, Any, Callable, Final

from uio import AnyReadableBuf, AnyWritableBuf

# noinspection SpellCheckingInspection
class BLE:
    """
   class BLE
   ---------
   """

    def __init__(self):
        """
       Returns the singleton BLE object.
      """
    @overload
    def active(self) -> bool:
        """
       Optionally changes the active state of the BLE radio, and returns the
       current state.
       
       The radio must be made active before using any other methods on this class.
      """
    @overload
    def active(self, active: bool, /) -> None:
        """
       Optionally changes the active state of the BLE radio, and returns the
       current state.
       
       The radio must be made active before using any other methods on this class.
      """
    @overload
    def config(self, param: str, /) -> Any:
        """
       Get or set configuration values of the BLE interface.  To get a value the
       parameter name should be quoted as a string, and just one parameter is
       queried at a time.  To set values use the keyword syntax, and one ore more
       parameter can be set at a time.
       
       Currently supported values are:
       
       - ``'mac'``: The current address in use, depending on the current address mode.
         This returns a tuple of ``(addr_type, addr)``.
       
         See :meth:`gatts_write <BLE.gatts_write>` for details about address type.
       
         This may only be queried while the interface is currently active.
       
       - ``'addr_mode'``: Sets the address mode. Values can be:
       
           * 0x00 - PUBLIC - Use the controller's public address.
           * 0x01 - RANDOM - Use a generated static address.
           * 0x02 - RPA - Use resolvable private addresses.
           * 0x03 - NRPA - Use non-resolvable private addresses.
       
         By default the interface mode will use a PUBLIC address if available, otherwise
         it will use a RANDOM address.
       
       - ``'gap_name'``: Get/set the GAP device name used by service 0x1800,
         characteristic 0x2a00.  This can be set at any time and changed multiple
         times.
       
       - ``'rxbuf'``: Get/set the size in bytes of the internal buffer used to store
         incoming events.  This buffer is global to the entire BLE driver and so
         handles incoming data for all events, including all characteristics.
         Increasing this allows better handling of bursty incoming data (for
         example scan results) and the ability to receive larger characteristic values.
       
       - ``'mtu'``: Get/set the MTU that will be used during a ATT MTU exchange. The
         resulting MTU will be the minimum of this and the remote device's MTU.
         ATT MTU exchange will not happen automatically (unless the remote device initiates
         it), and must be manually initiated with
         :meth:`gattc_exchange_mtu<BLE.gattc_exchange_mtu>`.
         Use the ``_IRQ_MTU_EXCHANGED`` event to discover the MTU for a given connection.
       
       - ``'bond'``: Sets whether bonding will be enabled during pairing. When
         enabled, pairing requests will set the "bond" flag and the keys will be stored
         by both devices.
       
       - ``'mitm'``: Sets whether MITM-protection is required for pairing.
       
       - ``'io'``: Sets the I/O capabilities of this device.
       
         Available options are::
       
           _IO_CAPABILITY_DISPLAY_ONLY = const(0)
           _IO_CAPABILITY_DISPLAY_YESNO = const(1)
           _IO_CAPABILITY_KEYBOARD_ONLY = const(2)
           _IO_CAPABILITY_NO_INPUT_OUTPUT = const(3)
           _IO_CAPABILITY_KEYBOARD_DISPLAY = const(4)
       
       - ``'le_secure'``: Sets whether "LE Secure" pairing is required. Default is
         false (i.e. allow "Legacy Pairing").
      """
    @overload
    def config(self, **kwargs) -> None:
        """
       Get or set configuration values of the BLE interface.  To get a value the
       parameter name should be quoted as a string, and just one parameter is
       queried at a time.  To set values use the keyword syntax, and one ore more
       parameter can be set at a time.
       
       Currently supported values are:
       
       - ``'mac'``: The current address in use, depending on the current address mode.
         This returns a tuple of ``(addr_type, addr)``.
       
         See :meth:`gatts_write <BLE.gatts_write>` for details about address type.
       
         This may only be queried while the interface is currently active.
       
       - ``'addr_mode'``: Sets the address mode. Values can be:
       
           * 0x00 - PUBLIC - Use the controller's public address.
           * 0x01 - RANDOM - Use a generated static address.
           * 0x02 - RPA - Use resolvable private addresses.
           * 0x03 - NRPA - Use non-resolvable private addresses.
       
         By default the interface mode will use a PUBLIC address if available, otherwise
         it will use a RANDOM address.
       
       - ``'gap_name'``: Get/set the GAP device name used by service 0x1800,
         characteristic 0x2a00.  This can be set at any time and changed multiple
         times.
       
       - ``'rxbuf'``: Get/set the size in bytes of the internal buffer used to store
         incoming events.  This buffer is global to the entire BLE driver and so
         handles incoming data for all events, including all characteristics.
         Increasing this allows better handling of bursty incoming data (for
         example scan results) and the ability to receive larger characteristic values.
       
       - ``'mtu'``: Get/set the MTU that will be used during a ATT MTU exchange. The
         resulting MTU will be the minimum of this and the remote device's MTU.
         ATT MTU exchange will not happen automatically (unless the remote device initiates
         it), and must be manually initiated with
         :meth:`gattc_exchange_mtu<BLE.gattc_exchange_mtu>`.
         Use the ``_IRQ_MTU_EXCHANGED`` event to discover the MTU for a given connection.
       
       - ``'bond'``: Sets whether bonding will be enabled during pairing. When
         enabled, pairing requests will set the "bond" flag and the keys will be stored
         by both devices.
       
       - ``'mitm'``: Sets whether MITM-protection is required for pairing.
       
       - ``'io'``: Sets the I/O capabilities of this device.
       
         Available options are::
       
           _IO_CAPABILITY_DISPLAY_ONLY = const(0)
           _IO_CAPABILITY_DISPLAY_YESNO = const(1)
           _IO_CAPABILITY_KEYBOARD_ONLY = const(2)
           _IO_CAPABILITY_NO_INPUT_OUTPUT = const(3)
           _IO_CAPABILITY_KEYBOARD_DISPLAY = const(4)
       
       - ``'le_secure'``: Sets whether "LE Secure" pairing is required. Default is
         false (i.e. allow "Legacy Pairing").
      """
    def irq(self, handler: Callable[[int, tuple[memoryview, ...]], Any], /) -> None:
        """
       Registers a callback for events from the BLE stack. The *handler* takes two
       arguments, ``event`` (which will be one of the codes below) and ``data``
       (which is an event-specific tuple of values).
       
       **Note:** As an optimisation to prevent unnecessary allocations, the ``addr``,
       ``adv_data``, ``char_data``, ``notify_data``, and ``uuid`` entries in the
       tuples are read-only memoryview instances pointing to :mod:`bluetooth`'s internal
       ringbuffer, and are only valid during the invocation of the IRQ handler
       function.  If your program needs to save one of these values to access after
       the IRQ handler has returned (e.g. by saving it in a class instance or global
       variable), then it needs to take a copy of the data, either by using ``bytes()``
       or ``bluetooth.UUID()``, like this::
       
           connected_addr = bytes(addr)  # equivalently: adv_data, char_data, or notify_data
           matched_uuid = bluetooth.UUID(uuid)
       
       For example, the IRQ handler for a scan result might inspect the ``adv_data``
       to decide if it's the correct device, and only then copy the address data to be
       used elsewhere in the program.  And to print data from within the IRQ handler,
       ``print(bytes(addr))`` will be needed.
       
       An event handler showing all possible events::
       
           def bt_irq(event, data):
               if event == _IRQ_CENTRAL_CONNECT:
                   # A central has connected to this peripheral.
                   conn_handle, addr_type, addr = data
               elif event == _IRQ_CENTRAL_DISCONNECT:
                   # A central has disconnected from this peripheral.
                   conn_handle, addr_type, addr = data
               elif event == _IRQ_GATTS_WRITE:
                   # A client has written to this characteristic or descriptor.
                   conn_handle, attr_handle = data
               elif event == _IRQ_GATTS_READ_REQUEST:
                   # A client has issued a read. Note: this is only supported on STM32.
                   # Return a non-zero integer to deny the read (see below), or zero (or None)
                   # to accept the read.
                   conn_handle, attr_handle = data
               elif event == _IRQ_SCAN_RESULT:
                   # A single scan result.
                   addr_type, addr, adv_type, rssi, adv_data = data
               elif event == _IRQ_SCAN_DONE:
                   # Scan duration finished or manually stopped.
                   pass
               elif event == _IRQ_PERIPHERAL_CONNECT:
                   # A successful gap_connect().
                   conn_handle, addr_type, addr = data
               elif event == _IRQ_PERIPHERAL_DISCONNECT:
                   # Connected peripheral has disconnected.
                   conn_handle, addr_type, addr = data
               elif event == _IRQ_GATTC_SERVICE_RESULT:
                   # Called for each service found by gattc_discover_services().
                   conn_handle, start_handle, end_handle, uuid = data
               elif event == _IRQ_GATTC_SERVICE_DONE:
                   # Called once service discovery is complete.
                   # Note: Status will be zero on success, implementation-specific value otherwise.
                   conn_handle, status = data
               elif event == _IRQ_GATTC_CHARACTERISTIC_RESULT:
                   # Called for each characteristic found by gattc_discover_services().
                   conn_handle, def_handle, value_handle, properties, uuid = data
               elif event == _IRQ_GATTC_CHARACTERISTIC_DONE:
                   # Called once service discovery is complete.
                   # Note: Status will be zero on success, implementation-specific value otherwise.
                   conn_handle, status = data
               elif event == _IRQ_GATTC_DESCRIPTOR_RESULT:
                   # Called for each descriptor found by gattc_discover_descriptors().
                   conn_handle, dsc_handle, uuid = data
               elif event == _IRQ_GATTC_DESCRIPTOR_DONE:
                   # Called once service discovery is complete.
                   # Note: Status will be zero on success, implementation-specific value otherwise.
                   conn_handle, status = data
               elif event == _IRQ_GATTC_READ_RESULT:
                   # A gattc_read() has completed.
                   conn_handle, value_handle, char_data = data
               elif event == _IRQ_GATTC_READ_DONE:
                   # A gattc_read() has completed.
                   # Note: The value_handle will be zero on btstack (but present on NimBLE).
                   # Note: Status will be zero on success, implementation-specific value otherwise.
                   conn_handle, value_handle, status = data
               elif event == _IRQ_GATTC_WRITE_DONE:
                   # A gattc_write() has completed.
                   # Note: The value_handle will be zero on btstack (but present on NimBLE).
                   # Note: Status will be zero on success, implementation-specific value otherwise.
                   conn_handle, value_handle, status = data
               elif event == _IRQ_GATTC_NOTIFY:
                   # A server has sent a notify request.
                   conn_handle, value_handle, notify_data = data
               elif event == _IRQ_GATTC_INDICATE:
                   # A server has sent an indicate request.
                   conn_handle, value_handle, notify_data = data
               elif event == _IRQ_GATTS_INDICATE_DONE:
                   # A client has acknowledged the indication.
                   # Note: Status will be zero on successful acknowledgment, implementation-specific value otherwise.
                   conn_handle, value_handle, status = data
               elif event == _IRQ_MTU_EXCHANGED:
                   # ATT MTU exchange complete (either initiated by us or the remote device).
                   conn_handle, mtu = data
               elif event == _IRQ_L2CAP_ACCEPT:
                   # A new channel has been accepted.
                   # Return a non-zero integer to reject the connection, or zero (or None) to accept.
                   conn_handle, cid, psm, our_mtu, peer_mtu = data
               elif event == _IRQ_L2CAP_CONNECT:
                   # A new channel is now connected (either as a result of connecting or accepting).
                   conn_handle, cid, psm, our_mtu, peer_mtu = data
               elif event == _IRQ_L2CAP_DISCONNECT:
                   # Existing channel has disconnected (status is zero), or a connection attempt failed (non-zero status).
                   conn_handle, cid, psm, status = data
               elif event == _IRQ_L2CAP_RECV:
                   # New data is available on the channel. Use l2cap_recvinto to read.
                   conn_handle, cid = data
               elif event == _IRQ_L2CAP_SEND_READY:
                   # A previous l2cap_send that returned False has now completed and the channel is ready to send again.
                   # If status is non-zero, then the transmit buffer overflowed and the application should re-send the data.
                   conn_handle, cid, status = data
               elif event == _IRQ_CONNECTION_UPDATE:
                   # The remote device has updated connection parameters.
                   conn_handle, conn_interval, conn_latency, supervision_timeout, status = data
               elif event == _IRQ_ENCRYPTION_UPDATE:
                   # The encryption state has changed (likely as a result of pairing or bonding).
                   conn_handle, encrypted, authenticated, bonded, key_size = data
               elif event == _IRQ_GET_SECRET:
                   # Return a stored secret.
                   # If key is None, return the index'th value of this sec_type.
                   # Otherwise return the corresponding value for this sec_type and key.
                   sec_type, index, key = data
                   return value
               elif event == _IRQ_SET_SECRET:
                   # Save a secret to the store for this sec_type and key.
                   sec_type, key, value = data
                   return True
               elif event == _IRQ_PASSKEY_ACTION:
                   # Respond to a passkey request during pairing.
                   # See gap_passkey() for details.
                   # action will be an action that is compatible with the configured "io" config.
                   # passkey will be non-zero if action is "numeric comparison".
                   conn_handle, action, passkey = data
       
       
       The event codes are::
       
       from micropython import const
       _IRQ_CENTRAL_CONNECT = const(1)
       _IRQ_CENTRAL_DISCONNECT = const(2)
       _IRQ_GATTS_WRITE = const(3)
       _IRQ_GATTS_READ_REQUEST = const(4)
       _IRQ_SCAN_RESULT = const(5)
       _IRQ_SCAN_DONE = const(6)
       _IRQ_PERIPHERAL_CONNECT = const(7)
       _IRQ_PERIPHERAL_DISCONNECT = const(8)
       _IRQ_GATTC_SERVICE_RESULT = const(9)
       _IRQ_GATTC_SERVICE_DONE = const(10)
       _IRQ_GATTC_CHARACTERISTIC_RESULT = const(11)
       _IRQ_GATTC_CHARACTERISTIC_DONE = const(12)
       _IRQ_GATTC_DESCRIPTOR_RESULT = const(13)
       _IRQ_GATTC_DESCRIPTOR_DONE = const(14)
       _IRQ_GATTC_READ_RESULT = const(15)
       _IRQ_GATTC_READ_DONE = const(16)
       _IRQ_GATTC_WRITE_DONE = const(17)
       _IRQ_GATTC_NOTIFY = const(18)
       _IRQ_GATTC_INDICATE = const(19)
       _IRQ_GATTS_INDICATE_DONE = const(20)
       _IRQ_MTU_EXCHANGED = const(21)
       _IRQ_L2CAP_ACCEPT = const(22)
       _IRQ_L2CAP_CONNECT = const(23)
       _IRQ_L2CAP_DISCONNECT = const(24)
       _IRQ_L2CAP_RECV = const(25)
       _IRQ_L2CAP_SEND_READY = const(26)
       _IRQ_CONNECTION_UPDATE = const(27)
       _IRQ_ENCRYPTION_UPDATE = const(28)
       _IRQ_GET_SECRET = const(29)
       _IRQ_SET_SECRET = const(30)
       
       For the ``_IRQ_GATTS_READ_REQUEST`` event, the available return codes are::
       
       _GATTS_NO_ERROR = const(0x00)
       _GATTS_ERROR_READ_NOT_PERMITTED = const(0x02)
       _GATTS_ERROR_WRITE_NOT_PERMITTED = const(0x03)
       _GATTS_ERROR_INSUFFICIENT_AUTHENTICATION = const(0x05)
       _GATTS_ERROR_INSUFFICIENT_AUTHORIZATION = const(0x08)
       _GATTS_ERROR_INSUFFICIENT_ENCRYPTION = const(0x0f)
       
       For the ``_IRQ_PASSKEY_ACTION`` event, the available actions are::
       
       _PASSKEY_ACTION_NONE = const(0)
       _PASSKEY_ACTION_INPUT = const(2)
       _PASSKEY_ACTION_DISPLAY = const(3)
       _PASSKEY_ACTION_NUMERIC_COMPARISON = const(4)
       
       In order to save space in the firmware, these constants are not included on the
       :mod:`bluetooth` module. Add the ones that you need from the list above to your
       program.
      """
    def gap_advertise(
        self,
        interval_us: int,
        adv_data: AnyReadableBuf | None = None,
        /,
        *,
        resp_data: AnyReadableBuf | None = None,
        connectable: bool = True,
    ) -> None:
        """
       Starts advertising at the specified interval (in **micro**\ seconds). This
       interval will be rounded down to the nearest 625us. To stop advertising, set
       *interval_us* to ``None``.
       
       *adv_data* and *resp_data* can be any type that implements the buffer
       protocol (e.g. ``bytes``, ``bytearray``, ``str``). *adv_data* is included
       in all broadcasts, and *resp_data* is send in reply to an active scan.
       
       **Note:** if *adv_data* (or *resp_data*) is ``None``, then the data passed
       to the previous call to ``gap_advertise`` will be re-used. This allows a
       broadcaster to resume advertising with just ``gap_advertise(interval_us)``.
       To clear the advertising payload pass an empty ``bytes``, i.e. ``b''``.
      """
    def gap_scan(
        self,
        duration_ms: int,
        interval_us: int = 1280000,
        window_us: int = 11250,
        active: bool = False,
        /,
    ) -> None:
        """
       Run a scan operation lasting for the specified duration (in **milli**\ seconds).
       
       To scan indefinitely, set *duration_ms* to ``0``.
       
       To stop scanning, set *duration_ms* to ``None``.
       
       Use *interval_us* and *window_us* to optionally configure the duty cycle.
       The scanner will run for *window_us* **micro**\ seconds every *interval_us*
       **micro**\ seconds for a total of *duration_ms* **milli**\ seconds. The default
       interval and window are 1.28 seconds and 11.25 milliseconds respectively
       (background scanning).
       
       For each scan result the ``_IRQ_SCAN_RESULT`` event will be raised, with event
       data ``(addr_type, addr, adv_type, rssi, adv_data)``.
       
       ``addr_type`` values indicate public or random addresses:
           * 0x00 - PUBLIC
           * 0x01 - RANDOM (either static, RPA, or NRPA, the type is encoded in the address itself)
       
       ``adv_type`` values correspond to the Bluetooth Specification:
       
           * 0x00 - ADV_IND - connectable and scannable undirected advertising
           * 0x01 - ADV_DIRECT_IND - connectable directed advertising
           * 0x02 - ADV_SCAN_IND - scannable undirected advertising
           * 0x03 - ADV_NONCONN_IND - non-connectable undirected advertising
           * 0x04 - SCAN_RSP - scan response
       
       ``active`` can be set ``True`` if you want to receive scan responses in the results.
       
       When scanning is stopped (either due to the duration finishing or when
       explicitly stopped), the ``_IRQ_SCAN_DONE`` event will be raised.
      """
    def gap_connect(
        self,
        addr_type: int,
        addr: bytes,
        scan_duration_ms: int = 2000,
        min_conn_interval_us: int | None = None,
        max_conn_interval_us: int | None = None,
        /,
    ) -> None:
        """
       Connect to a peripheral.
       
       See :meth:`gap_scan <BLE.gap_scan>` for details about address types.
       
       To cancel an outstanding connection attempt early, call
       ``gap_connect(None)``.
       
       On success, the ``_IRQ_PERIPHERAL_CONNECT`` event will be raised. If
       cancelling a connection attempt, the ``_IRQ_PERIPHERAL_DISCONNECT`` event
       will be raised.
       
       The device will wait up to *scan_duration_ms* to receive an advertising
       payload from the device.
       
       The connection interval can be configured in **micro**\ seconds using either
       or both of *min_conn_interval_us* and *max_conn_interval_us*. Otherwise a
       default interval will be chosen, typically between 30000 and 50000
       microseconds. A shorter interval will increase throughput, at the expense
       of power usage.
       
       
       
       Central Role
       ------------
       
       A central device can connect to peripherals that it has discovered using the observer role (see :meth:`gap_scan<BLE.gap_scan>`) or with a known address.
      """
    def gap_disconnect(self, conn_handle: memoryview, /) -> bool:
        """
       Disconnect the specified connection handle. This can either be a
       central that has connected to this device (if acting as a peripheral)
       or a peripheral that was previously connected to by this device (if acting
       as a central).
       
       On success, the ``_IRQ_PERIPHERAL_DISCONNECT`` or ``_IRQ_CENTRAL_DISCONNECT``
       event will be raised.
       
       Returns ``False`` if the connection handle wasn't connected, and ``True``
       otherwise.
       
       
       
       Central Role
       ------------
       
       A central device can connect to peripherals that it has discovered using the observer role (see :meth:`gap_scan<BLE.gap_scan>`) or with a known address.
       
       
       Peripheral Role
       ---------------
       
       A peripheral device is expected to send connectable advertisements (see
       :meth:`gap_advertise<BLE.gap_advertise>`). It will usually be acting as a GATT
       server, having first registered services and characteristics using
       :meth:`gatts_register_services<BLE.gatts_register_services>`.
       
       When a central connects, the ``_IRQ_CENTRAL_CONNECT`` event will be raised.
      """
    _Flag: Final = int
    _Descriptor: Final = tuple["UUID", _Flag]
    _Characteristic: Final = tuple["UUID", _Flag] | tuple[
        "UUID", _Flag, tuple[_Descriptor, ...]
    ]
    _Service: Final = tuple["UUID", tuple[_Characteristic, ...]]
    def gatts_register_services(
        self, services_definition: tuple[_Service, ...], /
    ) -> tuple[tuple[memoryview, ...], ...]:
        """
       Configures the server with the specified services, replacing any
       existing services.
       
       *services_definition* is a list of **services**, where each **service** is a
       two-element tuple containing a UUID and a list of **characteristics**.
       
       Each **characteristic** is a two-or-three-element tuple containing a UUID, a
       **flags** value, and optionally a list of *descriptors*.
       
       Each **descriptor** is a two-element tuple containing a UUID and a **flags**
       value.
       
       The **flags** are a bitwise-OR combination of the flags defined below. These
       set both the behaviour of the characteristic (or descriptor) as well as the
       security and privacy requirements.
       
       The return value is a list (one element per service) of tuples (each element
       is a value handle). Characteristics and descriptor handles are flattened
       into the same tuple, in the order that they are defined.
       
       The following example registers two services (Heart Rate, and Nordic UART)::
       
           HR_UUID = bluetooth.UUID(0x180D)
           HR_CHAR = (bluetooth.UUID(0x2A37), bluetooth.FLAG_READ | bluetooth.FLAG_NOTIFY,)
           HR_SERVICE = (HR_UUID, (HR_CHAR,),)
           UART_UUID = bluetooth.UUID('6E400001-B5A3-F393-E0A9-E50E24DCCA9E')
           UART_TX = (bluetooth.UUID('6E400003-B5A3-F393-E0A9-E50E24DCCA9E'), bluetooth.FLAG_READ | bluetooth.FLAG_NOTIFY,)
           UART_RX = (bluetooth.UUID('6E400002-B5A3-F393-E0A9-E50E24DCCA9E'), bluetooth.FLAG_WRITE,)
           UART_SERVICE = (UART_UUID, (UART_TX, UART_RX,),)
           SERVICES = (HR_SERVICE, UART_SERVICE,)
           ( (hr,), (tx, rx,), ) = bt.gatts_register_services(SERVICES)
       
       The three value handles (``hr``, ``tx``, ``rx``) can be used with
       :meth:`gatts_read <BLE.gatts_read>`, :meth:`gatts_write <BLE.gatts_write>`, :meth:`gatts_notify <BLE.gatts_notify>`, and
       :meth:`gatts_indicate <BLE.gatts_indicate>`.
       
       **Note:** Advertising must be stopped before registering services.
       
       Available flags for characteristics and descriptors are::
       
           from micropython import const
           _FLAG_BROADCAST = const(0x0001)
           _FLAG_READ = const(0x0002)
           _FLAG_WRITE_NO_RESPONSE = const(0x0004)
           _FLAG_WRITE = const(0x0008)
           _FLAG_NOTIFY = const(0x0010)
           _FLAG_INDICATE = const(0x0020)
           _FLAG_AUTHENTICATED_SIGNED_WRITE = const(0x0040)
       
           _FLAG_AUX_WRITE = const(0x0100)
           _FLAG_READ_ENCRYPTED = const(0x0200)
           _FLAG_READ_AUTHENTICATED = const(0x0400)
           _FLAG_READ_AUTHORIZED = const(0x0800)
           _FLAG_WRITE_ENCRYPTED = const(0x1000)
           _FLAG_WRITE_AUTHENTICATED = const(0x2000)
           _FLAG_WRITE_AUTHORIZED = const(0x4000)
       
       As for the IRQs above, any required constants should be added to your Python code.
       
       
       GATT Server
       -----------
       
       A GATT server has a set of registered services. Each service may contain
       characteristics, which each have a value. Characteristics can also contain
       descriptors, which themselves have values.
       
       These values are stored locally, and are accessed by their "value handle" which
       is generated during service registration. They can also be read from or written
       to by a remote client device. Additionally, a server can "notify" a
       characteristic to a connected client via a connection handle.
       
       A device in either central or peripheral roles may function as a GATT server,
       however in most cases it will be more common for a peripheral device to act
       as the server.
       
       Characteristics and descriptors have a default maximum size of 20 bytes.
       Anything written to them by a client will be truncated to this length. However,
       any local write will increase the maximum size, so if you want to allow larger
       writes from a client to a given characteristic, use
       :meth:`gatts_write<BLE.gatts_write>` after registration. e.g.
       ``gatts_write(char_handle, bytes(100))``.
      """
    def gatts_read(self, value_handle: memoryview, /) -> bytes:
        """
       Reads the local value for this handle (which has either been written by
       :meth:`gatts_write <BLE.gatts_write>` or by a remote client).
       
       
       GATT Server
       -----------
       
       A GATT server has a set of registered services. Each service may contain
       characteristics, which each have a value. Characteristics can also contain
       descriptors, which themselves have values.
       
       These values are stored locally, and are accessed by their "value handle" which
       is generated during service registration. They can also be read from or written
       to by a remote client device. Additionally, a server can "notify" a
       characteristic to a connected client via a connection handle.
       
       A device in either central or peripheral roles may function as a GATT server,
       however in most cases it will be more common for a peripheral device to act
       as the server.
       
       Characteristics and descriptors have a default maximum size of 20 bytes.
       Anything written to them by a client will be truncated to this length. However,
       any local write will increase the maximum size, so if you want to allow larger
       writes from a client to a given characteristic, use
       :meth:`gatts_write<BLE.gatts_write>` after registration. e.g.
       ``gatts_write(char_handle, bytes(100))``.
      """
    def gatts_write(
        self, value_handle: memoryview, data: bytes, send_update: bool = False, /
    ) -> None:
        """
       Writes the local value for this handle, which can be read by a client.
       
       If *send_update* is ``True``, then any subscribed clients will be notified
       (or indicated, depending on what they're subscribed to and which operations
       the characteristic supports) about this write.
       
       
       GATT Server
       -----------
       
       A GATT server has a set of registered services. Each service may contain
       characteristics, which each have a value. Characteristics can also contain
       descriptors, which themselves have values.
       
       These values are stored locally, and are accessed by their "value handle" which
       is generated during service registration. They can also be read from or written
       to by a remote client device. Additionally, a server can "notify" a
       characteristic to a connected client via a connection handle.
       
       A device in either central or peripheral roles may function as a GATT server,
       however in most cases it will be more common for a peripheral device to act
       as the server.
       
       Characteristics and descriptors have a default maximum size of 20 bytes.
       Anything written to them by a client will be truncated to this length. However,
       any local write will increase the maximum size, so if you want to allow larger
       writes from a client to a given characteristic, use
       :meth:`gatts_write<BLE.gatts_write>` after registration. e.g.
       ``gatts_write(char_handle, bytes(100))``.
      """
    def gatts_notify(self, value_handle: memoryview, data: bytes, /) -> None:
        """
       Sends a notification request to a connected client.
       
       If *data* is not ``None``, then that value is sent to the client as part of
       the notification. The local value will not be modified.
       
       Otherwise, if *data* is ``None``, then the current local value (as
       set with :meth:`gatts_write <BLE.gatts_write>`) will be sent.
       
       **Note:** The notification will be sent regardless of the subscription
       status of the client to this characteristic.
       
       
       GATT Server
       -----------
       
       A GATT server has a set of registered services. Each service may contain
       characteristics, which each have a value. Characteristics can also contain
       descriptors, which themselves have values.
       
       These values are stored locally, and are accessed by their "value handle" which
       is generated during service registration. They can also be read from or written
       to by a remote client device. Additionally, a server can "notify" a
       characteristic to a connected client via a connection handle.
       
       A device in either central or peripheral roles may function as a GATT server,
       however in most cases it will be more common for a peripheral device to act
       as the server.
       
       Characteristics and descriptors have a default maximum size of 20 bytes.
       Anything written to them by a client will be truncated to this length. However,
       any local write will increase the maximum size, so if you want to allow larger
       writes from a client to a given characteristic, use
       :meth:`gatts_write<BLE.gatts_write>` after registration. e.g.
       ``gatts_write(char_handle, bytes(100))``.
      """
    def gatts_indicate(
        self, conn_handle: memoryview, value_handle: memoryview, /
    ) -> None:
        """
       Sends an indication request containing the characteristic's current value to
       a connected client.
       
       On acknowledgment (or failure, e.g. timeout), the
       ``_IRQ_GATTS_INDICATE_DONE`` event will be raised.
       
       **Note:** The indication will be sent regardless of the subscription
       status of the client to this characteristic.
       
       
       GATT Server
       -----------
       
       A GATT server has a set of registered services. Each service may contain
       characteristics, which each have a value. Characteristics can also contain
       descriptors, which themselves have values.
       
       These values are stored locally, and are accessed by their "value handle" which
       is generated during service registration. They can also be read from or written
       to by a remote client device. Additionally, a server can "notify" a
       characteristic to a connected client via a connection handle.
       
       A device in either central or peripheral roles may function as a GATT server,
       however in most cases it will be more common for a peripheral device to act
       as the server.
       
       Characteristics and descriptors have a default maximum size of 20 bytes.
       Anything written to them by a client will be truncated to this length. However,
       any local write will increase the maximum size, so if you want to allow larger
       writes from a client to a given characteristic, use
       :meth:`gatts_write<BLE.gatts_write>` after registration. e.g.
       ``gatts_write(char_handle, bytes(100))``.
      """
    def gatts_set_buffer(
        self, conn_handle: memoryview, len: int, append: bool = False, /
    ) -> None:
        """
       Sets the internal buffer size for a value in bytes. This will limit the
       largest possible write that can be received. The default is 20.
       
       Setting *append* to ``True`` will make all remote writes append to, rather
       than replace, the current value. At most *len* bytes can be buffered in
       this way. When you use :meth:`gatts_read <BLE.gatts_read>`, the value will
       be cleared after reading. This feature is useful when implementing something
       like the Nordic UART Service.
       
       
       GATT Server
       -----------
       
       A GATT server has a set of registered services. Each service may contain
       characteristics, which each have a value. Characteristics can also contain
       descriptors, which themselves have values.
       
       These values are stored locally, and are accessed by their "value handle" which
       is generated during service registration. They can also be read from or written
       to by a remote client device. Additionally, a server can "notify" a
       characteristic to a connected client via a connection handle.
       
       A device in either central or peripheral roles may function as a GATT server,
       however in most cases it will be more common for a peripheral device to act
       as the server.
       
       Characteristics and descriptors have a default maximum size of 20 bytes.
       Anything written to them by a client will be truncated to this length. However,
       any local write will increase the maximum size, so if you want to allow larger
       writes from a client to a given characteristic, use
       :meth:`gatts_write<BLE.gatts_write>` after registration. e.g.
       ``gatts_write(char_handle, bytes(100))``.
      """
    def gattc_discover_services(
        self, conn_handle: memoryview, uuid: UUID | None = None, /
    ) -> None:
        """
       Query a connected server for its services.
       
       Optionally specify a service *uuid* to query for that service only.
       
       For each service discovered, the ``_IRQ_GATTC_SERVICE_RESULT`` event will
       be raised, followed by ``_IRQ_GATTC_SERVICE_DONE`` on completion.
       
       
       GATT Client
       -----------
       
       A GATT client can discover and read/write characteristics on a remote GATT server.
       
       It is more common for a central role device to act as the GATT client, however
       it's also possible for a peripheral to act as a client in order to discover
       information about the central that has connected to it (e.g. to read the
       device name from the device information service).
      """
    def gattc_discover_characteristics(
        self,
        conn_handle: memoryview,
        start_handle: int,
        end_handle: int,
        uuid: UUID | None = None,
        /,
    ) -> None:
        """
       Query a connected server for characteristics in the specified range.
       
       Optionally specify a characteristic *uuid* to query for that
       characteristic only.
       
       You can use ``start_handle=1``, ``end_handle=0xffff`` to search for a
       characteristic in any service.
       
       For each characteristic discovered, the ``_IRQ_GATTC_CHARACTERISTIC_RESULT``
       event will be raised, followed by ``_IRQ_GATTC_CHARACTERISTIC_DONE`` on completion.
       
       
       GATT Client
       -----------
       
       A GATT client can discover and read/write characteristics on a remote GATT server.
       
       It is more common for a central role device to act as the GATT client, however
       it's also possible for a peripheral to act as a client in order to discover
       information about the central that has connected to it (e.g. to read the
       device name from the device information service).
      """
    def gattc_discover_descriptors(
        self, conn_handle: memoryview, start_handle: int, end_handle: int, /
    ) -> None:
        """
       Query a connected server for descriptors in the specified range.
       
       For each descriptor discovered, the ``_IRQ_GATTC_DESCRIPTOR_RESULT`` event
       will be raised, followed by ``_IRQ_GATTC_DESCRIPTOR_DONE`` on completion.
       
       
       GATT Client
       -----------
       
       A GATT client can discover and read/write characteristics on a remote GATT server.
       
       It is more common for a central role device to act as the GATT client, however
       it's also possible for a peripheral to act as a client in order to discover
       information about the central that has connected to it (e.g. to read the
       device name from the device information service).
      """
    def gattc_read(self, conn_handle: memoryview, value_handle: memoryview, /) -> None:
        """
       Issue a remote read to a connected server for the specified
       characteristic or descriptor handle.
       
       When a value is available, the ``_IRQ_GATTC_READ_RESULT`` event will be
       raised. Additionally, the ``_IRQ_GATTC_READ_DONE`` will be raised.
       
       
       GATT Client
       -----------
       
       A GATT client can discover and read/write characteristics on a remote GATT server.
       
       It is more common for a central role device to act as the GATT client, however
       it's also possible for a peripheral to act as a client in order to discover
       information about the central that has connected to it (e.g. to read the
       device name from the device information service).
      """
    def gattc_write(
        self,
        conn_handle: memoryview,
        value_handle: memoryview,
        data: bytes,
        mode: int = 0,
        /,
    ) -> None:
        """
       Issue a remote write to a connected server for the specified
       characteristic or descriptor handle.
       
       The argument *mode* specifies the write behaviour, with the currently
       supported values being:
       
           * ``mode=0`` (default) is a write-without-response: the write will
             be sent to the remote server but no confirmation will be
             returned, and no event will be raised.
           * ``mode=1`` is a write-with-response: the remote server is
             requested to send a response/acknowledgement that it received the
             data.
       
       If a response is received from the remote server the
       ``_IRQ_GATTC_WRITE_DONE`` event will be raised.
       
       
       GATT Client
       -----------
       
       A GATT client can discover and read/write characteristics on a remote GATT server.
       
       It is more common for a central role device to act as the GATT client, however
       it's also possible for a peripheral to act as a client in order to discover
       information about the central that has connected to it (e.g. to read the
       device name from the device information service).
      """
    def gattc_exchange_mtu(self, conn_handle: memoryview, /) -> None:
        """
       Initiate MTU exchange with a connected server, using the preferred MTU
       set using ``BLE.config(mtu=value)``.
       
       The ``_IRQ_MTU_EXCHANGED`` event will be raised when MTU exchange
       completes.
       
       **Note:** MTU exchange is typically initiated by the central. When using
       the BlueKitchen stack in the central role, it does not support a remote
       peripheral initiating the MTU exchange. NimBLE works for both roles.
       
       
       
       GATT Client
       -----------
       
       A GATT client can discover and read/write characteristics on a remote GATT server.
       
       It is more common for a central role device to act as the GATT client, however
       it's also possible for a peripheral to act as a client in order to discover
       information about the central that has connected to it (e.g. to read the
       device name from the device information service).
      """
    def l2cap_listen(self, psm: memoryview, mtu: memoryview, /) -> None:
        """
       Start listening for incoming L2CAP channel requests on the specified *psm*
       with the local MTU set to *mtu*.
       
       When a remote device initiates a connection, the ``_IRQ_L2CAP_ACCEPT``
       event will be raised, which gives the listening server a chance to reject
       the incoming connection (by returning a non-zero integer).
       
       Once the connection is accepted, the ``_IRQ_L2CAP_CONNECT`` event will be
       raised, allowing the server to obtain the channel id (CID) and the local and
       remote MTU.
       
       **Note:** It is not currently possible to stop listening.
       
       
       L2CAP connection-oriented-channels
       ----------------------------------
       
          This feature allows for socket-like data exchange between two BLE devices.
          Once the devices are connected via GAP, either device can listen for the
          other to connect on a numeric PSM (Protocol/Service Multiplexer).
       
          **Note:** This is currently only supported when using the NimBLE stack on
          STM32 and Unix (not ESP32). Only one L2CAP channel may be active at a given
          time (i.e. you cannot connect while listening).
       
          Active L2CAP channels are identified by the connection handle that they were
          established on and a CID (channel ID).
       
          Connection-oriented channels have built-in credit-based flow control. Unlike
          ATT, where devices negotiate a shared MTU, both the listening and connecting
          devices each set an independent MTU which limits the maximum amount of
          outstanding data that the remote device can send before it is fully consumed
          in :meth:`l2cap_recvinto <BLE.l2cap_recvinto>`.
      """
    def l2cap_connect(
        self, conn_handle: memoryview, psm: memoryview, mtu: memoryview, /
    ) -> None:
        """
       Connect to a listening peer on the specified *psm* with local MTU set to *mtu*.
       
       On successful connection, the the ``_IRQ_L2CAP_CONNECT`` event will be
       raised, allowing the client to obtain the CID and the local and remote (peer) MTU.
       
       An unsuccessful connection will raise the ``_IRQ_L2CAP_DISCONNECT`` event
       with a non-zero status.
       
       
       L2CAP connection-oriented-channels
       ----------------------------------
       
          This feature allows for socket-like data exchange between two BLE devices.
          Once the devices are connected via GAP, either device can listen for the
          other to connect on a numeric PSM (Protocol/Service Multiplexer).
       
          **Note:** This is currently only supported when using the NimBLE stack on
          STM32 and Unix (not ESP32). Only one L2CAP channel may be active at a given
          time (i.e. you cannot connect while listening).
       
          Active L2CAP channels are identified by the connection handle that they were
          established on and a CID (channel ID).
       
          Connection-oriented channels have built-in credit-based flow control. Unlike
          ATT, where devices negotiate a shared MTU, both the listening and connecting
          devices each set an independent MTU which limits the maximum amount of
          outstanding data that the remote device can send before it is fully consumed
          in :meth:`l2cap_recvinto <BLE.l2cap_recvinto>`.
      """
    def l2cap_disconnect(self, conn_handle: memoryview, cid: memoryview, /) -> None:
        """
       Disconnect an active L2CAP channel with the specified *conn_handle* and
       *cid*.
       
       
       L2CAP connection-oriented-channels
       ----------------------------------
       
          This feature allows for socket-like data exchange between two BLE devices.
          Once the devices are connected via GAP, either device can listen for the
          other to connect on a numeric PSM (Protocol/Service Multiplexer).
       
          **Note:** This is currently only supported when using the NimBLE stack on
          STM32 and Unix (not ESP32). Only one L2CAP channel may be active at a given
          time (i.e. you cannot connect while listening).
       
          Active L2CAP channels are identified by the connection handle that they were
          established on and a CID (channel ID).
       
          Connection-oriented channels have built-in credit-based flow control. Unlike
          ATT, where devices negotiate a shared MTU, both the listening and connecting
          devices each set an independent MTU which limits the maximum amount of
          outstanding data that the remote device can send before it is fully consumed
          in :meth:`l2cap_recvinto <BLE.l2cap_recvinto>`.
      """
    def l2cap_send(self, conn_handle: memoryview, cid: memoryview, /) -> None:
        """
       Send the specified *buf* (which must support the buffer protocol) on the
       L2CAP channel identified by *conn_handle* and *cid*.
       
       The specified buffer cannot be larger than the remote (peer) MTU, and no
       more than twice the size of the local MTU.
       
       This will return ``False`` if the channel is now "stalled", which means that
       :meth:`l2cap_send <BLE.l2cap_send>` must not be called again until the
       ``_IRQ_L2CAP_SEND_READY`` event is received (which will happen when the
       remote device grants more credits, typically after it has received and
       processed the data).
       
       
       L2CAP connection-oriented-channels
       ----------------------------------
       
          This feature allows for socket-like data exchange between two BLE devices.
          Once the devices are connected via GAP, either device can listen for the
          other to connect on a numeric PSM (Protocol/Service Multiplexer).
       
          **Note:** This is currently only supported when using the NimBLE stack on
          STM32 and Unix (not ESP32). Only one L2CAP channel may be active at a given
          time (i.e. you cannot connect while listening).
       
          Active L2CAP channels are identified by the connection handle that they were
          established on and a CID (channel ID).
       
          Connection-oriented channels have built-in credit-based flow control. Unlike
          ATT, where devices negotiate a shared MTU, both the listening and connecting
          devices each set an independent MTU which limits the maximum amount of
          outstanding data that the remote device can send before it is fully consumed
          in :meth:`l2cap_recvinto <BLE.l2cap_recvinto>`.
      """
    def l2cap_recvinto(
        self, conn_handle: memoryview, cid: memoryview, buf: AnyWritableBuf | None, /
    ) -> int:
        """
       Receive data from the specified *conn_handle* and *cid* into the provided
       *buf* (which must support the buffer protocol, e.g. bytearray or
       memoryview).
       
       Returns the number of bytes read from the channel.
       
       If *buf* is None, then returns the number of bytes available.
       
       **Note:** After receiving the ``_IRQ_L2CAP_RECV`` event, the application should
       continue calling :meth:`l2cap_recvinto <BLE.l2cap_recvinto>` until no more
       bytes are available in the receive buffer (typically up to the size of the
       remote (peer) MTU).
       
       Until the receive buffer is empty, the remote device will not be granted
       more channel credits and will be unable to send any more data.
       
       
       
       L2CAP connection-oriented-channels
       ----------------------------------
       
          This feature allows for socket-like data exchange between two BLE devices.
          Once the devices are connected via GAP, either device can listen for the
          other to connect on a numeric PSM (Protocol/Service Multiplexer).
       
          **Note:** This is currently only supported when using the NimBLE stack on
          STM32 and Unix (not ESP32). Only one L2CAP channel may be active at a given
          time (i.e. you cannot connect while listening).
       
          Active L2CAP channels are identified by the connection handle that they were
          established on and a CID (channel ID).
       
          Connection-oriented channels have built-in credit-based flow control. Unlike
          ATT, where devices negotiate a shared MTU, both the listening and connecting
          devices each set an independent MTU which limits the maximum amount of
          outstanding data that the remote device can send before it is fully consumed
          in :meth:`l2cap_recvinto <BLE.l2cap_recvinto>`.
      """
    def gap_pair(self, conn_handle: memoryview, /) -> None:
        """
       Initiate pairing with the remote device.
       
       Before calling this, ensure that the ``io``, ``mitm``, ``le_secure``, and
       ``bond`` configuration options are set (via :meth:`config<BLE.config>`).
       
       On successful pairing, the ``_IRQ_ENCRYPTION_UPDATE`` event will be raised.
       
       
       Pairing and bonding
       -------------------
       
          Pairing allows a connection to be encrypted and authenticated via exchange
          of secrets (with optional MITM protection via passkey authentication).
       
          Bonding is the process of storing those secrets into non-volatile storage.
          When bonded, a device is able to resolve a resolvable private address (RPA)
          from another device based on the stored identity resolving key (IRK).
          To support bonding, an application must implement the ``_IRQ_GET_SECRET``
          and ``_IRQ_SET_SECRET`` events.
       
          **Note:** This is currently only supported when using the NimBLE stack on
          STM32 and Unix (not ESP32).
      """
    def gap_passkey(
        self, conn_handle: memoryview, action: int, passkey: int, /
    ) -> None:
        """
       Respond to a ``_IRQ_PASSKEY_ACTION`` event for the specified *conn_handle*
       and *action*.
       
       The *passkey* is a numeric value and will depend on on the
       *action* (which will depend on what I/O capability has been set):
       
           * When the *action* is ``_PASSKEY_ACTION_INPUT``, then the application should
             prompt the user to enter the passkey that is shown on the remote device.
           * When the *action* is ``_PASSKEY_ACTION_DISPLAY``, then the application should
             generate a random 6-digit passkey and show it to the user.
           * When the *action* is ``_PASSKEY_ACTION_NUMERIC_COMPARISON``, then the application
             should show the passkey that was provided in the ``_IRQ_PASSKEY_ACTION`` event
             and then respond with either ``0`` (cancel pairing), or ``1`` (accept pairing).
       
       
       
       Pairing and bonding
       -------------------
       
          Pairing allows a connection to be encrypted and authenticated via exchange
          of secrets (with optional MITM protection via passkey authentication).
       
          Bonding is the process of storing those secrets into non-volatile storage.
          When bonded, a device is able to resolve a resolvable private address (RPA)
          from another device based on the stored identity resolving key (IRK).
          To support bonding, an application must implement the ``_IRQ_GET_SECRET``
          and ``_IRQ_SET_SECRET`` events.
       
          **Note:** This is currently only supported when using the NimBLE stack on
          STM32 and Unix (not ESP32).
      """

class UUID:
    """
   class UUID
   ----------
   """

    def __init__(self, value: int | str, /):
        """
       Creates a UUID instance with the specified **value**.
       
       The **value** can be either:
       
       - A 16-bit integer. e.g. ``0x2908``.
       - A 128-bit UUID string. e.g. ``'6E400001-B5A3-F393-E0A9-E50E24DCCA9E'``.
      """



"""
The `socketpool` module provides sockets through a pool. The pools themselves
act like CPython's `socket` module.

For more information about the `socket` module, see the CPython documentation:
https://docs.python.org/3/library/socket.html
"""

from __future__ import annotations

from typing import Optional, Tuple

import socketpool
import wifi
from circuitpython_typing import ReadableBuffer, WriteableBuffer

class Socket:
    """TCP, UDP and RAW socket. Cannot be created directly. Instead, call
    `SocketPool.socket()`.

    Provides a subset of CPython's `socket.socket` API. It only implements the versions of
    recv that do not allocate bytes objects."""

    def __hash__(self) -> int:
        """Returns a hash for the Socket."""
        ...
    def __enter__(self) -> Socket:
        """No-op used by Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically closes the Socket when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    def accept(self) -> Tuple[Socket, Tuple[str, int]]:
        """Accept a connection on a listening socket of type SOCK_STREAM,
        creating a new socket of type SOCK_STREAM.
        Returns a tuple of (new_socket, remote_address)"""
    def bind(self, address: Tuple[str, int]) -> None:
        """Bind a socket to an address

        :param ~tuple address: tuple of (remote_address, remote_port)"""
        ...
    def close(self) -> None:
        """Closes this Socket and makes its resources available to its SocketPool."""
    def connect(self, address: Tuple[str, int]) -> None:
        """Connect a socket to a remote address

        :param ~tuple address: tuple of (remote_address, remote_port)"""
        ...
    def listen(self, backlog: int) -> None:
        """Set socket to listen for incoming connections

        :param ~int backlog: length of backlog queue for waiting connections"""
        ...
    def recvfrom_into(self, buffer: WriteableBuffer) -> Tuple[int, Tuple[str, int]]:
        """Reads some bytes from a remote address.

        Returns a tuple containing
        * the number of bytes received into the given buffer
        * a remote_address, which is a tuple of ip address and port number

        :param object buffer: buffer to read into"""
        ...
    def recv_into(self, buffer: WriteableBuffer, bufsize: int) -> int:
        """Reads some bytes from the connected remote address, writing
        into the provided buffer. If bufsize <= len(buffer) is given,
        a maximum of bufsize bytes will be read into the buffer. If no
        valid value is given for bufsize, the default is the length of
        the given buffer.

        Suits sockets of type SOCK_STREAM
        Returns an int of number of bytes read.

        :param bytearray buffer: buffer to receive into
        :param int bufsize: optionally, a maximum number of bytes to read."""
        ...
    def send(self, bytes: ReadableBuffer) -> int:
        """Send some bytes to the connected remote address.
        Suits sockets of type SOCK_STREAM

        :param ~bytes bytes: some bytes to send"""
        ...
    def sendall(self, bytes: ReadableBuffer) -> None:
        """Send some bytes to the connected remote address.
        Suits sockets of type SOCK_STREAM

        This calls send() repeatedly until all the data is sent or an error
        occurs. If an error occurs, it's impossible to tell how much data
        has been sent.

        :param ~bytes bytes: some bytes to send"""
        ...
    def sendto(self, bytes: ReadableBuffer, address: Tuple[str, int]) -> int:
        """Send some bytes to a specific address.
        Suits sockets of type SOCK_DGRAM

        :param ~bytes bytes: some bytes to send
        :param ~tuple address: tuple of (remote_address, remote_port)"""
        ...
    def setblocking(self, flag: bool) -> Optional[int]:
        """Set the blocking behaviour of this socket.

        :param ~bool flag: False means non-blocking, True means block indefinitely."""
        ...
    def setsockopt(self, level: int, optname: int, value: int) -> None:
        """Sets socket options"""
        ...
    def settimeout(self, value: int) -> None:
        """Set the timeout value for this socket.

        :param ~int value: timeout in seconds.  0 means non-blocking.  None means block indefinitely.
        """
        ...

class SocketPool:
    """A pool of socket resources available for the given radio. Only one
    SocketPool can be created for each radio.

    SocketPool should be used in place of CPython's socket which provides
    a pool of sockets provided by the underlying OS.
    """

    def __init__(self, radio: wifi.Radio) -> None:
        """Create a new SocketPool object for the provided radio

        :param wifi.Radio radio: The (connected) network hardware to associate
            with this SocketPool; currently, this will always be the object
            returned by :py:attr:`wifi.radio`
        """
        ...

    class gaierror(OSError):
        """Errors raised by getaddrinfo"""

    AF_INET: int
    AF_INET6: int

    SOCK_STREAM: int
    SOCK_DGRAM: int
    SOCK_RAW: int
    EAI_NONAME: int

    TCP_NODELAY: int

    IPPROTO_TCP: int
    def socket(
        self, family: int = AF_INET, type: int = SOCK_STREAM
    ) -> socketpool.Socket:
        """Create a new socket

        :param ~int family: AF_INET or AF_INET6
        :param ~int type: SOCK_STREAM, SOCK_DGRAM or SOCK_RAW

        The ``proto`` (protocol) and ``fileno`` arguments available in ``socket.socket()``
        in CPython are not supported.
        """
        ...
    def getaddrinfo(
        self,
        host: str,
        port: int,
        family: int = 0,
        type: int = 0,
        proto: int = 0,
        flags: int = 0,
    ) -> Tuple[int, int, int, str, Tuple[str, int]]:
        """Gets the address information for a hostname and port

        Returns the appropriate family, socket type, socket protocol and
        address information to call socket.socket() and socket.connect() with,
        as a tuple."""
        ...



"""Dualbank Module

The `dualbank` module adds ability to update and switch between the
two identical app partitions, which can contain different firmware versions.

Having two partitions enables rollback functionality.

The two partitions are defined as the boot partition and the next-update partition.
Calling `dualbank.flash()` writes the next-update partition.

After the next-update partition is written a validation check is performed
and on a successful validation this partition is set as the boot partition.
On next reset, firmware will be loaded from this partition.

Use cases:
    * Can be used for ``OTA`` Over-The-Air updates.
    * Can be used for ``dual-boot`` of different firmware versions or platforms.

.. note::

    Boards with flash ``=2MB``:
        This module is unavailable as the flash is only large enough for one app partition.

    Boards with flash ``>2MB``:
        This module is enabled/disabled at runtime based on whether the ``CIRCUITPY`` drive
        is extended or not. See `storage.erase_filesystem()` for more information.

.. code-block:: python

    import dualbank

    dualbank.flash(buffer, offset)
    dualbank.switch()
"""

from __future__ import annotations

from circuitpython_typing import ReadableBuffer

...

def flash(buffer: ReadableBuffer, offset: int = 0) -> None:
    """Writes one of the two app partitions at the given offset.

    This can be called multiple times when flashing the firmware in smaller chunks.

    :param ReadableBuffer buffer: The entire firmware or a partial chunk.
    :param int offset: Start writing at this offset in the app partition.
    """
    ...

def switch() -> None:
    """Switches to the next-update partition.

    On next reset, firmware will be loaded from the partition just switched over to.
    """
    ...



import os.path
import sys
from logging import getLogger
from typing import List, Optional

import turcar
from turcar.plugins.micropython.bare_metal_backend import (
    BareMetalMicroPythonBackend,
    launch_bare_metal_backend,
)

logger = getLogger("turcar.plugins.prime_inventor.prime_inventor_back")


class PrimeInventorMicroPythonBackend(BareMetalMicroPythonBackend):
    def _get_sys_path_for_analysis(self) -> Optional[List[str]]:
        return [
            os.path.join(os.path.dirname(__file__), "api_stubs"),
        ] + super()._get_sys_path_for_analysis()


if __name__ == "__main__":
    launch_bare_metal_backend(PrimeInventorMicroPythonBackend)



from pybricks.parameters import Port


class UltrasonicSensor:
    """
    LEGO® MINDSTORMS® EV3 Ultrasonic Sensor.

    Args:
        port (Port): Port to which the sensor is connected.
    """

    def __init__(self, port: Port):
        if port == Port.A or port == Port.B or port == Port.C or port == Port.D:
            raise ValueError("Sensors must use Port S1, S2, S3, or S4.")

    def distance(self, silent: bool = False) -> int:
        """
        Measures the distance between the sensor and an object using ultrasonic sound waves.

        Args:
            silent (bool): Choose True to turn the sensor off after measuring the distance. This reduces interference with other ultrasonic sensors. If you do this too frequently, the sensor can freeze. If this happens, unplug it and plug it back in.

        Returns:
            Distance in millimeters.
        """
        return 0

    def presence(self) -> bool:
        """
        Check for the presence of other ultrasonic sensors by detecting ultrasonic sounds. 

        If the other ultrasonic sensor is operating in silent mode, you can only detect the presence of that sensor while it is taking a measurement.

        Returns:
            True if ultrasonic sounds are detected, False if not.
        """
        return False



import re
import time
import tkinter as tk
from logging import getLogger

from turcar import get_workbench, ui_utils
from turcar.languages import tr
from turcar.ui_utils import ems_to_pixels

logger = getLogger(__name__)

INFO_TEXT = "---"


class TodoView(ui_utils.TreeFrame):
    def __init__(self, master):
        ui_utils.TreeFrame.__init__(
            self,
            master,
            columns=("line_no", "todo_text"),
            displaycolumns=(0, 1),
        )

        self._current_code_view = None
        self._current_source = None

        self.tree.bind("<<TreeviewSelect>>", self._on_click, True)
        self.tree.bind("<Map>", self._update, True)

        get_workbench().bind("WorkbenchReady", self._update, True)
        get_workbench().bind("Save", self._update, True)
        get_workbench().bind("SaveAs", self._update, True)

        get_workbench().bind_class("Text", "<Double-Button-1>", self._update, True)
        get_workbench().bind_class("Text", "<<NewLine>>", self._update, True)

        get_workbench().get_editor_notebook().bind("<<NotebookTabChanged>>", self._update, True)
        get_workbench().bind_class("EditorCodeViewText", "<<TextChange>>", self._text_change, True)

        self.tree.column("line_no", width=ems_to_pixels(4), anchor=tk.W)
        self.tree.column("todo_text", width=ems_to_pixels(100), anchor=tk.W)

        self.tree.heading("line_no", text=tr("Line"), anchor=tk.W)
        self.tree.heading("todo_text", text=tr("Info"), anchor=tk.W)

        self.tree["show"] = ["headings"]

        self._update(None)

    def _last_op_delta(self):
        last_op = self._current_code_view.text.get_last_operation_time()
        now = time.time()
        return now - last_op

    def _text_change(self, event):
        if self._current_code_view is None:
            return

        if not hasattr(self._current_code_view, "_update_already_scheduled"):
            self._current_code_view._update_already_scheduled = False

        if self._current_code_view._update_already_scheduled:
            return

        if self._last_op_delta() < 0.3:
            self._current_code_view._update_already_scheduled = True

            def delay_update():
                if self._last_op_delta() < 0.3:
                    # still typing
                    self._current_code_view.text.after(100, delay_update)
                    return

                self._current_code_view._update_already_scheduled = False
                self._update(None)

            self._current_code_view.text.after_idle(delay_update)

        else:
            self._update(event)

    def _update(self, event):
        if not self.winfo_ismapped():
            return

        editor = get_workbench().get_editor_notebook().get_current_editor()

        if editor is None:
            self._current_code_view = None
            self._current_source = None
            return

        new_codeview = editor.get_code_view()
        new_source = new_codeview.get_content()

        if self._current_code_view == new_codeview and self._current_source == new_source:
            return

        self.clear()

        self._current_code_view = new_codeview
        self._current_source = new_source

        # todo support of other file types and introducing comment tags

        r_ex = r"^.*((#.*[\t ]*(TODO|BUG|FIXME|ERROR|NOTE|REMARK)([:\t ]*))(.*))$"
        r_match = re.compile(r_ex, re.IGNORECASE | re.MULTILINE)

        line_no = 0
        for line in new_source.splitlines():
            line_no += 1
            matches = r_match.finditer(line)
            if matches:
                for m in matches:
                    todo_text = m.groups()[0]
                    self.tree.insert("", "end", values=(line_no, todo_text))

        if len(self.tree.get_children()) == 0:
            # todo enhance the regex so that a todo within quotes is not shown in the list
            # low prio
            self.tree.insert("", "end", values=(INFO_TEXT, tr("No line marked with #todo found")))

    def clear(self):
        self.tree.delete(*self.tree.get_children())

    def _on_click(self, event):
        if self._current_code_view is None:
            return

        iid = self.tree.focus()

        if iid != "":
            values = self.tree.item(iid)["values"]
            line_no = values[0]
            if line_no and line_no != INFO_TEXT:
                editor = get_workbench().get_editor_notebook().get_current_editor()
                editor.select_line(line_no)


def load_plugin() -> None:
    get_workbench().add_view(TodoView, tr("TODO"), "s")



"""USB Human Interface Device

The `usb_hid` module allows you to output data as a HID device."""

from __future__ import annotations

from typing import Optional, Sequence, Tuple

from circuitpython_typing import ReadableBuffer

devices: Tuple[Device, ...]
"""Tuple of all active HID device interfaces.
The default set of devices is ``Device.KEYBOARD, Device.MOUSE, Device.CONSUMER_CONTROL``,
On boards where `usb_hid` is disabled by default, `devices` is an empty tuple.

If a boot device is enabled by `usb_hid.enable()`, *and* the host has requested a boot device,
the `devices` tuple is **replaced** when ``code.py`` starts with a single-element tuple
containing a `Device` that describes the boot device chosen (keyboard or mouse).
The request for a boot device overrides any other HID devices.
"""

def disable() -> None:
    """Do not present any USB HID devices to the host computer.
    Can be called in ``boot.py``, before USB is connected.
    The HID composite device is normally enabled by default,
    but on some boards with limited endpoints, including STM32F4,
    it is disabled by default. You must turn off another USB device such
    as `usb_cdc` or `storage` to free up endpoints for use by `usb_hid`.
    """

def enable(devices: Optional[Sequence[Device]], boot_device: int = 0) -> None:
    """Specify which USB HID devices that will be available.
    Can be called in ``boot.py``, before USB is connected.

    :param Sequence devices: `Device` objects.
      If `devices` is empty, HID is disabled. The order of the ``Devices``
      may matter to the host. For instance, for MacOS, put the mouse device
      before any Gamepad or Digitizer HID device or else it will not work.
    :param int boot_device: If non-zero, inform the host that support for a
      a boot HID device is available.
      If ``boot_device=1``, a boot keyboard is available.
      If ``boot_device=2``, a boot mouse is available. No other values are allowed.
      See below.

    If you enable too many devices at once, you will run out of USB endpoints.
    The number of available endpoints varies by microcontroller.
    CircuitPython will go into safe mode after running ``boot.py`` to inform you if
    not enough endpoints are available.

    **Boot Devices**

    Boot devices implement a fixed, predefined report descriptor, defined in
    https://www.usb.org/sites/default/files/hid1_12.pdf, Appendix B. A USB host
    can request to use the boot device if the USB device says it is available.
    Usually only a BIOS or other kind of limited-functionality
    host needs boot keyboard support.

    For example, to make a boot keyboard available, you can use this code::

      usb_hid.enable((Device.KEYBOARD), boot_device=1)  # 1 for a keyboard

    If the host requests the boot keyboard, the report descriptor provided by `Device.KEYBOARD`
    will be ignored, and the predefined report descriptor will be used.
    But if the host does not request the boot keyboard,
    the descriptor provided by `Device.KEYBOARD` will be used.

    The HID boot device must usually be the first or only device presented by CircuitPython.
    The HID device will be USB interface number 0.
    To make sure it is the first device, disable other USB devices, including CDC and MSC (CIRCUITPY).
    If you specify a non-zero ``boot_device``, and it is not the first device, CircuitPython
    will enter safe mode to report this error.
    """
    ...

def get_boot_device() -> int:
    """
    :return: the boot device requested by the host, if any.
      Returns 0 if the host did not request a boot device, or if `usb_hid.enable()`
      was called with ``boot_device=0``, the default, which disables boot device support.
      If the host did request a boot device,
      returns the value of ``boot_device`` set in `usb_hid.enable()`:
      ``1`` for a boot keyboard, or ``2`` for boot mouse.
      However, the standard devices provided by CircuitPython, `Device.KEYBOARD` and `Device.MOUSE`,
      describe reports that match the boot device reports, so you don't need to check this
      if you are using those devices.
    :rtype int:
    """

class Device:
    """HID Device specification"""

    def __init__(
        self,
        *,
        report_descriptor: ReadableBuffer,
        usage_page: int,
        usage: int,
        report_ids: Sequence[int],
        in_report_lengths: Sequence[int],
        out_report_lengths: Sequence[int],
    ) -> None:
        """Create a description of a USB HID device. The actual device is created when you
        pass a `Device` to `usb_hid.enable()`.

        :param ReadableBuffer report_descriptor: The USB HID Report descriptor bytes. The descriptor is not
          not verified for correctness; it is up to you to make sure it is not malformed.
        :param int usage_page: The Usage Page value from the descriptor. Must match what is in the descriptor.
        :param int usage: The Usage value from the descriptor. Must match what is in the descriptor.
        :param Sequence[int] report_ids: Sequence of report ids used by the descriptor.
          If the ``report_descriptor`` does not specify any report IDs, use ``(0,)``.
        :param Sequence[int] in_report_lengths: Sequence of sizes in bytes of the HID reports sent to the host.
          The sizes are in order of the ``report_ids``.
          Use a size of ``0`` for a report that is not an IN report.
          "IN" is with respect to the host.
        :param int out_report_lengths: Sequence of sizes in bytes of the HID reports received from the host.
          The sizes are in order of the ``report_ids``.
          Use a size of ``0`` for a report that is not an OUT report.
          "OUT" is with respect to the host.

        ``report_ids``, ``in_report_lengths``, and ``out_report_lengths`` must all have the
        same number of elements.

        Here is an example of a `Device` with a descriptor that specifies two report IDs, 3 and 4.
        Report ID 3 sends an IN report of length 5, and receives an OUT report of length 6.
        Report ID 4 sends an IN report of length 2, and does not receive an OUT report::

            device = usb_hid.Device(
                descriptor=b"...",         # Omitted for brevity.
                report_ids=(3, 4),
                in_report_lengths=(5, 2),
                out_report_lengths=(6, 0),
            )
        """
        ...
    KEYBOARD: Device
    """Standard keyboard device supporting keycodes 0x00-0xDD, modifiers 0xE-0xE7, and five LED indicators.
    Uses Report ID 1 for its IN and OUT reports.
    """

    MOUSE: Device
    """Standard mouse device supporting five mouse buttons, X and Y relative movements from -127 to 127
    in each report, and a relative mouse wheel change from -127 to 127 in each report.
    Uses Report ID 2 for its IN report.
    """

    CONSUMER_CONTROL: Device
    """Consumer Control device supporting sent values from 1-652, with no rollover.
    Uses Report ID 3 for its IN report."""
    def send_report(
        self, report: ReadableBuffer, report_id: Optional[int] = None
    ) -> None:
        """Send an HID report. If the device descriptor specifies zero or one report id's,
        you can supply `None` (the default) as the value of ``report_id``.
        Otherwise you must specify which report id to use when sending the report.
        """
        ...
    def get_last_received_report(
        self, report_id: Optional[int] = None
    ) -> Optional[bytes]:
        """Get the last received HID OUT or feature report for the given report ID.
        The report ID may be omitted if there is no report ID, or only one report ID.
        Return `None` if nothing received. After returning a report, subsequent calls
        will return `None` until next report is received.
        """
        ...
    usage_page: int
    """The device usage page identifier, which designates a category of device. (read-only)"""
    usage: int
    """The device usage identifier, which designates a specific kind of device. (read-only)

    For example, Keyboard is 0x06 within the generic desktop usage page 0x01.
    Mouse is 0x02 within the same usage page."""



"""SAMD implementation settings"""

from __future__ import annotations

from typing import Union

""":mod:`samd.clock` --- samd clock names
--------------------------------------------------------

.. module:: samd.clock
  :synopsis: samd clock names
  :platform: SAMD21

References to clocks as named by the microcontroller"""

class Clock:
    """Identifies a clock on the microcontroller.

    They are fixed by the hardware so they cannot be constructed on demand. Instead, use
    ``samd.clock`` to reference the desired clock."""

    enabled: bool
    """Is the clock enabled? (read-only)"""
    parent: Union[Clock, None]
    """Clock parent. (read-only)"""
    frequency: int
    """Clock frequency in Herz. (read-only)"""
    calibration: int
    """Clock calibration. Not all clocks can be calibrated."""



import os.path
import sys
from logging import getLogger
from typing import List, Optional

# make sure turcar folder is in sys.path (relevant in dev)
turcar_container = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
if turcar_container not in sys.path:
    sys.path.insert(0, turcar_container)


from turcar.plugins.micropython.bare_metal_backend import (
    BareMetalMicroPythonBackend,
    launch_bare_metal_backend,
)

# Can't use __name__, because it will be "__main__"
logger = getLogger("turcar.plugins.micropython.cirpy_backend")

_ENTER_REPL_PHRASES = [
    "Press any key to enter the REPL. Use CTRL-D to reload.",
    "Appuyez sur n'importe quelle touche pour utiliser le REPL. Utilisez CTRL-D pour relancer.",
    "Presiona cualquier tecla para entrar al REPL. Usa CTRL-D para recargar.",
    "Drücke eine beliebige Taste um REPL zu betreten. Drücke STRG-D zum neuladen.",
    "Druk een willekeurige toets om de REPL te starten. Gebruik CTRL+D om te herstarten.",
    "àn rèn hé jiàn jìn rù REPL. shǐ yòng CTRL-D zhòng xīn jiā zǎi ."
    "Tekan sembarang tombol untuk masuk ke REPL. Tekan CTRL-D untuk memuat ulang.",
    "Pressione qualquer tecla para entrar no REPL. Use CTRL-D para recarregar.",
    "Tryck på valfri tangent för att gå in i REPL. Använd CTRL-D för att ladda om.",
    "Нажмите любую клавишу чтобы зайти в REPL. Используйте CTRL-D для перезагрузки.",
]

_AUTO_RELOAD_PHRASES = [
    "Auto-reload is on. Simply save files over USB to run them or enter REPL to disable.",
    "Auto-chargement activé. Copiez ou sauvegardez les fichiers via USB pour les lancer ou démarrez le REPL pour le désactiver.",
    "Auto-reload habilitado. Simplemente guarda los archivos via USB para ejecutarlos o entra al REPL para desabilitarlos.",
    "Automatisches Neuladen ist aktiv. Speichere Dateien über USB um sie auszuführen oder verbinde dich mit der REPL zum Deaktivieren.",
    "L'auto-reload è attivo. Salva i file su USB per eseguirli o entra nel REPL per disabilitarlo.",
    "Auto-reload be on. Put yer files on USB to weigh anchor, er' bring'er about t' the REPL t' scuttle.",
    "Auto-herlaad staat aan. Sla bestanden simpelweg op over USB om uit te voeren of start REPL om uit te schakelen.",
    "Ang awtomatikong pag re-reload ay ON. i-save lamang ang mga files sa USB para patakbuhin sila o pasukin ang REPL para i-disable ito.",
    "Zìdòng chóngxīn jiāzài. Zhǐ xū tōngguò USB bǎocún wénjiàn lái yùnxíng tāmen huò shūrù REPL jìnyòng.",
    "Auto-reload aktif. Silahkan simpan data-data (files) melalui USB untuk menjalankannya atau masuk ke REPL untukmenonaktifkan.",
    "Samo-przeładowywanie włączone. Po prostu zapisz pliki przez USB aby je uruchomić, albo wejdź w konsolę aby wyłączyć.",
    "O recarregamento automático está ativo. Simplesmente salve os arquivos via USB para executá-los ou digite REPL para desativar.",
    "Autoladdning är på. Spara filer via USB för att köra dem eller ange REPL för att inaktivera.",
    "Автоматическая перезагрузка включена. Просто сохрани файл по USB или зайди в REPL чтобы отключить.",
]


class CircuitPythonBackend(BareMetalMicroPythonBackend):
    def _clear_repl(self):
        """
        CP runs code.py after soft-reboot even in raw repl.
        At the same time, it re-initializes VM and hardware just by switching
        between raw and friendly REPL (tested in CP 6.3 and 7.1)
        """
        logger.info("Creating fresh REPL for CP")
        self._ensure_normal_mode()
        self._ensure_raw_mode()

    def _transform_output(self, data, stream_name):
        # Any keypress wouldn't work in turcar's shell
        for phrase in _ENTER_REPL_PHRASES + _AUTO_RELOAD_PHRASES:
            data = data.replace(phrase + "\r\n", "").replace(phrase + "\n", "")

        return data

    def _output_warrants_interrupt(self, data):
        data = data.strip()
        for phrase in _ENTER_REPL_PHRASES:
            if data.endswith(phrase.encode("utf-8")):
                return True

        return False

    def _get_sys_path_for_analysis(self) -> Optional[List[str]]:
        return [
            os.path.join(os.path.dirname(__file__), "api_stubs"),
        ]


if __name__ == "__main__":
    launch_bare_metal_backend(CircuitPythonBackend)



"""Raw memory map access

The `memorymap` module allows you to read and write memory addresses in the
address space seen from the processor running CircuitPython. It is usually
the physical address space.
"""

from __future__ import annotations

from typing import overload

from circuitpython_typing import ReadableBuffer

class AddressRange:
    r"""Presents a range of addresses as a bytearray.

    The addresses may access memory or memory mapped peripherals.

    Some address ranges may be protected by CircuitPython to prevent errors.
    An exception will be raised when constructing an AddressRange for an
    invalid or protected address.

    Multiple AddressRanges may overlap. There is no "claiming" of addresses.

    Example usage on ESP32-S2::

       import memorymap
       rtc_slow_mem = memorymap.AddressRange(start=0x50000000, length=0x2000)
       rtc_slow_mem[0:3] = b"\xcc\x10\x00"
    """
    def __init__(self, *, start, length) -> None:
        """Constructs an address range starting at ``start`` and ending at
        ``start + length``. An exception will be raised if any of the
        addresses are invalid or protected."""
        ...
    def __bool__(self) -> bool: ...
    def __len__(self) -> int:
        """Return the length. This is used by (`len`)"""
        ...
    @overload
    def __getitem__(self, index: slice) -> bytearray: ...
    @overload
    def __getitem__(self, index: int) -> int:
        """Returns the value(s) at the given index.

        1, 2, 4 and 8 byte aligned reads will be done in one transaction.
        All others may use multiple transactions."""
        ...
    @overload
    def __setitem__(self, index: slice, value: ReadableBuffer) -> None: ...
    @overload
    def __setitem__(self, index: int, value: int) -> None:
        """Set the value(s) at the given index.

        1, 2, 4 and 8 byte aligned writes will be done in one transaction.
        All others may use multiple transactions."""
        ...



# -*- coding: utf-8 -*-

import re

from turcar import get_runner, get_workbench, ui_utils
from turcar.codeview import CodeViewText

cell_regex = re.compile(r"(^|\n)(# ?%%|##|# In\[\d+\]:)[^\n]*", re.MULTILINE)  # @UndefinedVariable


def update_editor_cells(event):
    text = event.widget

    if not getattr(text, "cell_tags_configured", False):
        text.tag_configure("CURRENT_CELL", borderwidth=1, relief="groove", background="LightYellow")
        text.tag_configure("CELL_HEADER", font="BoldEditorFont", foreground="#665843")

        text.tag_lower("CELL_HEADER")
        text.tag_lower("CURRENT_CELL")
        text.cell_tags_configured = True

    text.tag_remove("CURRENT_CELL", "0.1", "end")
    text.tag_remove("CELL_HEADER", "0.1", "end")
    source = text.get("1.0", "end")
    cells = []
    prev_marker = 0
    for match in cell_regex.finditer(source):
        if match.start() == 0:
            this_marker = match.start()
        else:
            this_marker = match.start() + 1

        cell_start_index = text.index("1.0+%dc" % prev_marker)
        header_end_index = text.index("1.0+%dc" % match.end())
        cell_end_index = text.index("1.0+%dc" % this_marker)
        text.tag_add("CELL_HEADER", cell_end_index, header_end_index)
        cells.append((cell_start_index, cell_end_index))

        prev_marker = this_marker

    if prev_marker != 0:
        cells.append((text.index("1.0+%dc" % prev_marker), "end"))

    # if get_workbench().focus_get() == text:
    # It's nice to have cell highlighted even when focus
    # is elsewhere ? This would act as kind of bookmark.

    for start_index, end_index in cells:
        if text.compare(start_index, "<=", "insert") and text.compare(end_index, ">", "insert"):
            text.tag_add("CURRENT_CELL", start_index, end_index)
            break


def _submit_code(code):
    lines = code.splitlines()

    # remove starting comments
    while len(lines) > 0 and lines[0].strip().startswith("#"):
        lines = lines[1:]

    # remove starting empty lines
    while len(lines) > 0 and lines[0].strip() == "":
        lines = lines[1:]

    # remove trailing empty lines
    while len(lines) > 0 and lines[-1].strip() == "":
        lines = lines[:-1]

    if len(lines) > 0:
        code = "\n".join(lines) + "\n"
        # if code is function definition/last line start with whitespace
        # end function definition with a second endline
        if re.match(r"^[ \t].*", lines[-1]) is not None:
            code += "\n"
        shell = get_workbench().show_view("ShellView", False)
        shell.submit_python_code(code)


def _patch_perform_return():
    original_perform_return = CodeViewText.perform_return

    def _patched_perform_return(self, event):
        text = event.widget
        ranges = text.tag_ranges("CURRENT_CELL")

        if len(ranges) == 2 and (
            ui_utils.shift_is_pressed(event) or ui_utils.control_is_pressed(event)
        ):
            if run_enabled():
                code = text.get(ranges[0], ranges[1]).strip()
                _submit_code(code)

                if ui_utils.shift_is_pressed(event):
                    # advance to next cell
                    text.mark_set("insert", ranges[1])

            return "break"
        else:
            return original_perform_return(self, event)

    CodeViewText.perform_return = _patched_perform_return


def _patch_intercept_mark():
    """Need to make cursor wider when in first column. Otherwise
    the border of the cell box makes it hard to notice the cursor.

    NB! Need to be careful with setting text["insertwidth"]!
    My first straightforward solution caused unexplainable
    infinite loop of insertions and deletions in the text
    (insert a line and a word, select that word and then do Ctrl-Z).

    Looks like this solution is safe, but I don't dare to include
    it in the main code.

    UPDATE: not safe. Select and delete a block of lines. Write a new
    line and do Ctrl-Z"""

    original_intercept_mark = CodeViewText.intercept_mark

    def _patched_intercept_mark(self, *args):
        if args[:2] == ("set", "insert") and args[2].endswith(".0"):
            self.set_insertwidth(3)
        else:
            self.set_insertwidth(2)

        original_intercept_mark(self, *args)

    CodeViewText.intercept_mark = _patched_intercept_mark


def dummy(event=None):
    "This is dummy method"


def run_selection(event=None):
    widget = get_workbench().focus_get()
    if isinstance(widget, CodeViewText):
        text = widget

        if text.has_selection():
            code = text.get("sel.first", "sel.last")
        else:
            code = text.get("insert linestart", "insert lineend")

        # move cursor to next row
        row, col = map(int, text.index("insert").split("."))
        text.mark_set("insert", "{}.{}".format(row + 1, col))

        _submit_code(code)


def run_enabled():
    widget = get_workbench().focus_get()
    return isinstance(widget, CodeViewText) and get_runner().is_waiting_toplevel_command()


def _load_plugin():
    wb = get_workbench()
    wb.bind_class("EditorCodeViewText", "<<CursorMove>>", update_editor_cells, True)
    wb.bind_class("EditorCodeViewText", "<<TextChange>>", update_editor_cells, True)
    wb.bind_class("EditorCodeViewText", "<FocusIn>", update_editor_cells, True)
    wb.bind_class("EditorCodeViewText", "<FocusOut>", update_editor_cells, True)

    _patch_perform_return()
    # TODO: try changing insertwidth in keyup/mouseup events
    # _patch_intercept_mark() # Still causes freezes

    wb.add_command(
        "run_cell",
        "run",
        ("Run cell"),
        handler=dummy,  # actual handler is in the patch
        default_sequence="<Control-Return>",
        tester=run_enabled,
        group=11,
    )

    wb.add_command(
        "run_cell_and_advance",
        "run",
        ("Run cell and advance"),
        handler=dummy,  # actual handler is in the patch
        default_sequence="<Shift-Return>",
        tester=run_enabled,
        group=11,
    )

    wb.add_command(
        "run_selection",
        "run",
        ("Run selection or current line"),
        handler=run_selection,
        default_sequence="<F9>",
        tester=run_enabled,
        group=11,
    )



"""`rainbowio` module.

Provides the `colorwheel()` function."""

from __future__ import annotations

def colorwheel(n: float) -> int:
    """C implementation of the common colorwheel() function found in many examples.
    Returns the colorwheel RGB value as an integer value for n (usable in neopixel and dotstar).
    """
    ...



"""
time related functions

Descriptions taken from
`https://raw.githubusercontent.com/micropython/micropython/master/docs/library/time.rst`, etc.
=====================================

.. module:: time
   :synopsis: time related functions

|see_cpython_module| :mod:`python:time`.

The ``time`` module provides functions for getting the current time and date,
measuring time intervals, and for delays.

**Time Epoch**: Unix port uses standard for POSIX systems epoch of
1970-01-01 00:00:00 UTC. However, embedded ports use epoch of
2000-01-01 00:00:00 UTC.

**Maintaining actual calendar date/time**: This requires a
Real Time Clock (RTC). On systems with underlying OS (including some
RTOS), an RTC may be implicit. Setting and maintaining actual calendar
time is responsibility of OS/RTOS and is done outside of MicroPython,
it just uses OS API to query date/time. On baremetal ports however
system time depends on ``machine.RTC()`` object. The current calendar time
may be set using ``machine.RTC().datetime(tuple)`` function, and maintained
by following means:

* By a backup battery (which may be an additional, optional component for
  a particular board).
* Using networked time protocol (requires setup by a port/user).
* Set manually by a user on each power-up (many boards then maintain
  RTC time across hard resets, though some may require setting it again
  in such case).

If actual calendar time is not maintained with a system/MicroPython RTC,
functions below which require reference to current absolute time may
behave not as expected.
"""

__author__ = "Howard C Lovatt"
__copyright__ = "Howard C Lovatt, 2020 onwards."
__license__ = "MIT https://opensource.org/licenses/MIT (as used by MicroPython)."
__version__ = "7.1.0"  # Version set by https://github.com/hlovatt/tag2ver

from typing import Final, Optional, Tuple, TypeVar

class _TicksMs: ...
class _TicksUs: ...
class _TicksCPU: ...

_Ticks: Final = TypeVar("_Ticks", _TicksMs, _TicksUs, _TicksCPU, int)


def sleep(seconds: float, /) -> None:
    """
   Sleep for the given number of seconds. Some boards may accept *seconds* as a
   floating-point number to sleep for a fractional number of seconds. Note that
   other boards may not accept a floating-point argument, for compatibility with
   them use `sleep_ms()` and `sleep_us()` functions.
   """

def sleep_ms(ms: int, /) -> None:
    """
   Delay for given number of milliseconds, should be positive or 0.

   This function will delay for at least the given number of milliseconds, but
   may take longer than that if other processing must take place, for example
   interrupt handlers or other threads.  Passing in 0 for *ms* will still allow
   this other processing to occur.  Use `sleep_us()` for more precise delays.
   """

def sleep_us(us: int, /) -> None:
    """
   Delay for given number of microseconds, should be positive or 0.

   This function attempts to provide an accurate delay of at least *us*
   microseconds, but it may take longer if the system has other higher priority
   processing to perform.
   """

def ticks_ms() -> _TicksMs:
    """
    Returns an increasing millisecond counter with an arbitrary reference point, that
    wraps around after some value.

    The wrap-around value is not explicitly exposed, but we will
    refer to it as *TICKS_MAX* to simplify discussion. Period of the values is
    *TICKS_PERIOD = TICKS_MAX + 1*. *TICKS_PERIOD* is guaranteed to be a power of
    two, but otherwise may differ from port to port. The same period value is used
    for all of `ticks_ms()`, `ticks_us()`, `ticks_cpu()` functions (for
    simplicity). Thus, these functions will return a value in range [*0* ..
    *TICKS_MAX*], inclusive, total *TICKS_PERIOD* values. Note that only
    non-negative values are used. For the most part, you should treat values returned
    by these functions as opaque. The only operations available for them are
    `ticks_diff()` and `ticks_add()` functions described below.

    Note: Performing standard mathematical operations (+, -) or relational
    operators (<, <=, >, >=) directly on these value will lead to invalid
    result. Performing mathematical operations and then passing their results
    as arguments to `ticks_diff()` or `ticks_add()` will also lead to
    invalid results from the latter functions.
   """

def ticks_us() -> _TicksUs:
    """
   Just like `ticks_ms()` above, but in microseconds.
   """


def ticks_add(ticks: _Ticks, delta: int, /) -> _Ticks:
    """
   Offset ticks value by a given number, which can be either positive or negative.
   Given a *ticks* value, this function allows to calculate ticks value *delta*
   ticks before or after it, following modular-arithmetic definition of tick values
   (see `ticks_ms()` above). *ticks* parameter must be a direct result of call
   to `ticks_ms()`, `ticks_us()`, or `ticks_cpu()` functions (or from previous
   call to `ticks_add()`). However, *delta* can be an arbitrary integer number
   or numeric expression. `ticks_add()` is useful for calculating deadlines for
   events/tasks. (Note: you must use `ticks_diff()` function to work with
   deadlines.)

   Examples::

        # Find out what ticks value there was 100ms ago
        print(ticks_add(time.ticks_ms(), -100))

        # Calculate deadline for operation and test for it
        deadline = ticks_add(time.ticks_ms(), 200)
        while ticks_diff(deadline, time.ticks_ms()) > 0:
            do_a_little_of_something()

        # Find out TICKS_MAX used by this port
        print(ticks_add(0, -1))
   """

def ticks_diff(ticks1: _Ticks, ticks2: _Ticks, /) -> int:
    """
   Measure ticks difference between values returned from `ticks_ms()`, `ticks_us()`,
   or `ticks_cpu()` functions, as a signed value which may wrap around.

   The argument order is the same as for subtraction
   operator, ``ticks_diff(ticks1, ticks2)`` has the same meaning as ``ticks1 - ticks2``.
   However, values returned by `ticks_ms()`, etc. functions may wrap around, so
   directly using subtraction on them will produce incorrect result. That is why
   `ticks_diff()` is needed, it implements modular (or more specifically, ring)
   arithmetics to produce correct result even for wrap-around values (as long as they not
   too distant inbetween, see below). The function returns **signed** value in the range
   [*-TICKS_PERIOD/2* .. *TICKS_PERIOD/2-1*] (that's a typical range definition for
   two's-complement signed binary integers). If the result is negative, it means that
   *ticks1* occurred earlier in time than *ticks2*. Otherwise, it means that
   *ticks1* occurred after *ticks2*. This holds **only** if *ticks1* and *ticks2*
   are apart from each other for no more than *TICKS_PERIOD/2-1* ticks. If that does
   not hold, incorrect result will be returned. Specifically, if two tick values are
   apart for *TICKS_PERIOD/2-1* ticks, that value will be returned by the function.
   However, if *TICKS_PERIOD/2* of real-time ticks has passed between them, the
   function will return *-TICKS_PERIOD/2* instead, i.e. result value will wrap around
   to the negative range of possible values.

   Informal rationale of the constraints above: Suppose you are locked in a room with no
   means to monitor passing of time except a standard 12-notch clock. Then if you look at
   dial-plate now, and don't look again for another 13 hours (e.g., if you fall for a
   long sleep), then once you finally look again, it may seem to you that only 1 hour
   has passed. To avoid this mistake, just look at the clock regularly. Your application
   should do the same. "Too long sleep" metaphor also maps directly to application
   behaviour: don't let your application run any single task for too long. Run tasks
   in steps, and do time-keeping inbetween.

   `ticks_diff()` is designed to accommodate various usage patterns, among them:

   * Polling with timeout. In this case, the order of events is known, and you will deal
     only with positive results of `ticks_diff()`::

        # Wait for GPIO pin to be asserted, but at most 500us
        start = time.ticks_us()
        while pin.value() == 0:
            if time.ticks_diff(time.ticks_us(), start) > 500:
                raise TimeoutError

   * Scheduling events. In this case, `ticks_diff()` result may be negative
     if an event is overdue::

        # This code snippet is not optimized
        now = time.ticks_ms()
        scheduled_time = task.scheduled_time()
        if ticks_diff(scheduled_time, now) > 0:
            print("Too early, let's nap")
            sleep_ms(ticks_diff(scheduled_time, now))
            task.run()
        elif ticks_diff(scheduled_time, now) == 0:
            print("Right at time!")
            task.run()
        elif ticks_diff(scheduled_time, now) < 0:
            print("Oops, running late, tell task to run faster!")
            task.run(run_faster=true)

   Note: Do not pass `time()` values to `ticks_diff()`, you should use
   normal mathematical operations on them. But note that `time()` may (and will)
   also overflow. This is known as https://en.wikipedia.org/wiki/Year_2038_problem .
   """




__all__ = ["ColorSensor", "GyroSensor", "InfraredSensor", "Motor", "TouchSensor", "UltrasonicSensor"]

from .__stub.__colorsensor import ColorSensor
from .__stub.__gyrosensor import GyroSensor
from .__stub.__infraredsensor import InfraredSensor
from .__stub.__motor import Motor
from .__stub.__touchsensor import TouchSensor
from .__stub.__ultrasonicsensor import UltrasonicSensor


"""Atexit Module

This module defines functions to register and unregister cleanup functions.
Functions thus registered are automatically executed upon normal vm termination.

These functions are run in the reverse order in which they were registered;
if you register ``A``, ``B``, and ``C``, they will be run in the order ``C``, ``B``, ``A``.

|see_cpython_module| :mod:`cpython:atexit`.
"""

from __future__ import annotations

from typing import Any, Callable, Optional

...

def register(
    func: Callable[..., Any], *args: Optional[Any], **kwargs: Optional[Any]
) -> Callable[..., Any]:
    """Register func as a function to be executed at termination.

    Any optional arguments that are to be passed to func must be passed as arguments to `register()`.
    It is possible to register the same function and arguments more than once.

    At normal program termination (for instance, if `sys.exit()` is called or the vm execution completes),
    all functions registered are called in last in, first out order.

    If an exception is raised during execution of the exit handler,
    a traceback is printed (unless `SystemExit` is raised) and the execution stops.

    This function returns func, which makes it possible to use it as a decorator.

    """
    ...

def unregister(func: Callable[..., Any]) -> None:
    """Remove func from the list of functions to be run at termination.

    `unregister()` silently does nothing if func was not previously registered. If func has been registered more than once,
    every occurrence of that function in the atexit call stack will be removed.

    """
    ...



# -*- coding: utf-8 -*-

"""
This file is run by CPythonProxy

(Why separate file for launching? I want to have clean global scope
in toplevel __main__ module (because that's where user scripts run), but backend's global scope
is far from clean.
I could also do python -c "from backend import MainCPythonBackend: MainCPythonBackend().mainloop()",
but looks like this gives relative __file__-s on imported modules.)
"""
import ast

# NB! This module can be also imported (when querying its path for uploading)
if __name__ == "__main__":
    import ast
    import os.path
    import sys

    # make sure turcar folder is in sys.path during startup
    turcar_container = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
    if turcar_container not in sys.path:
        # We're running with non-turcar interpreter or env
        sys.path.insert(0, turcar_container)
        using_temp_augmented_sys_path = True
    else:
        using_temp_augmented_sys_path = False

    if sys.platform == "darwin":
        try:
            os.getcwd()
        except Exception:
            print(
                "\nNB! Potential problems detected, see\nhttps://github.com/turcar/turcar/wiki/MacOSX#catalina\n",
                file=sys.stderr,
            )

    if not sys.version_info > (3, 8):
        print(
            "This version of turcar only supports Python 3.8 and later.\n"
            + "Choose another interpreter from Tools => Options => Interpreter",
            file=sys.stderr,
        )
        sys.exit(1)
    import turcar
    from turcar import report_time

    report_time("Before importing MainCPythonBackend")
    from turcar.common import PROCESS_ACK
    from turcar.plugins.cpython_backend.cp_back import MainCPythonBackend

    turcar.prepare_turcar_user_dir()
    turcar.configure_backend_logging()
    print(PROCESS_ACK)

    if using_temp_augmented_sys_path:
        # Don't make turcar container available for user programs.
        # NB! Need to do it before constructing the backend as it would clean the main scope and the flag would be gone.
        from logging import getLogger

        assert sys.path[0] == turcar_container
        getLogger(__name__).info("Removing temporary %r from sys.path", turcar_container)
        del sys.path[0]

    target_cwd = sys.argv[1]
    options = ast.literal_eval(sys.argv[2])
    report_time("Before constructing backend")
    # Don't introduce new variables after constructing the backend, as it cleaned the main scope
    MainCPythonBackend(target_cwd, options).mainloop()



import ast
import binascii
import errno
import os
import re
import struct
import time
from abc import ABC
from logging import getLogger
from textwrap import dedent
from typing import Any, Callable, Dict, List, Optional, Tuple

from pipkin import UserError
from pipkin.adapters import BaseAdapter
from pipkin.common import CommunicationError, ManagementError, ProtocolError
from pipkin.connection import MicroPythonConnection, ReadingTimeoutError
from pipkin.serial_connection import SerialConnection
from pipkin.util import starts_with_continuation_byte

logger = getLogger(__name__)

RAW_MODE_CMD = b"\x01"
NORMAL_MODE_CMD = b"\x02"
INTERRUPT_CMD = b"\x03"
SOFT_REBOOT_CMD = b"\x04"
PASTE_MODE_CMD = b"\x05"
PASTE_MODE_LINE_PREFIX = b"=== "

PASTE_SUBMIT_MODE = "paste"
RAW_PASTE_SUBMIT_MODE = "raw_paste"
RAW_SUBMIT_MODE = "raw"

RAW_PASTE_COMMAND = b"\x05A\x01"
RAW_PASTE_CONFIRMATION = b"R\x01"
RAW_PASTE_CONTINUE = b"\x01"

MGMT_VALUE_START = b"<pipkin>"
MGMT_VALUE_END = b"</pipkin>"

# How many seconds to wait for something that should appear quickly.
# In other words -- how long to wait with reporting a protocol error
# (hoping that the required piece is still coming)
WAIT_OR_CRASH_TIMEOUT = 5

FIRST_RAW_PROMPT = b"raw REPL; CTRL-B to exit\r\n>"

RAW_PROMPT = b">"


WEBREPL_REQ_S = "<2sBBQLH64s"
WEBREPL_PUT_FILE = 1
WEBREPL_GET_FILE = 2


EOT = b"\x04"
NORMAL_PROMPT = b">>> "
LF = b"\n"
OK = b"OK"
ESC = b"\x1b"
ST = b"\x1b\\"

ENCODING = "utf-8"
TRACEBACK_MARKER = b"Traceback (most recent call last):"

OutputConsumer = Callable[[str, str], None]


class BareMetalAdapter(BaseAdapter, ABC):
    def __init__(
        self,
        connection: MicroPythonConnection,
        submit_mode: Optional[str] = None,
        write_block_size: Optional[int] = None,
        write_block_delay: Optional[float] = None,
    ):
        super().__init__()
        self._connection = connection
        (
            self._submit_mode,
            self._write_block_size,
            self._write_block_delay,
        ) = self._infer_submit_parameters(submit_mode, write_block_size, write_block_delay)
        self._last_prompt: Optional[bytes] = None

        self._interrupt_to_prompt()
        self._prepare_helper()
        self._builtin_modules = self._fetch_builtin_modules()
        logger.debug("Builtin modules: %r", self._builtin_modules)

    def get_dir_sep(self) -> str:
        return "/"

    def _infer_submit_parameters(
        self,
        submit_mode: Optional[str] = None,
        write_block_size: Optional[int] = None,
        write_block_delay: Optional[float] = None,
    ) -> Tuple[str, int, float]:
        if submit_mode is None:
            submit_mode = RAW_PASTE_SUBMIT_MODE

        if write_block_size is None:
            write_block_size = 255

        if write_block_delay is None:
            if submit_mode == RAW_SUBMIT_MODE:
                write_block_delay = 0.01
            else:
                write_block_delay = 0.0

        return submit_mode, write_block_size, write_block_delay

    def _fetch_builtin_modules(self) -> List[str]:
        script = "__pipkin_helper.builtins.help('modules')"
        out, err = self._execute_and_capture_output(script)
        if err or not out:
            logger.warning("Could not query builtin modules")
            return []

        modules_str_lines = out.strip().splitlines()

        last_line = modules_str_lines[-1].strip()
        if last_line.count(" ") > 0 and "  " not in last_line and "\t" not in last_line:
            # probably something like "plus any modules on the filesystem"
            # (can be in different languages)
            modules_str_lines = modules_str_lines[:-1]

        modules_str = (
            " ".join(modules_str_lines)
            .replace("/__init__", "")
            .replace("__main__", "")
            .replace("/", ".")
        )

        return modules_str.split()

    def _interrupt_to_prompt(self) -> None:
        # It's safer to thoroughly interrupt before poking with RAW_MODE_CMD
        # as Pico may get stuck otherwise
        # https://github.com/micropython/micropython/issues/7867
        interventions = [(INTERRUPT_CMD, 0.1), (INTERRUPT_CMD, 0.1), (RAW_MODE_CMD, 0.1)]

        for cmd, timeout in interventions:
            self._write(cmd)
            try:
                self._log_output_until_active_prompt(timeout=timeout)
                break
            except TimeoutError as e:
                logger.debug(
                    "Could not get prompt with intervention %r and timeout %r. Read bytes: %r",
                    cmd,
                    timeout,
                    getattr(e, "read_bytes", "?"),
                )
                # Try again as long as there are interventions left
        else:
            raise CommunicationError("Could not get raw REPL")

    def _log_output(self, data: bytes, stream: str = "stdout") -> None:
        logger.debug("read %s: %r", stream, data)

    def _prepare_helper(self) -> None:
        script = (
            dedent(
                """
            class __pipkin_helper:
                import builtins
                try:
                    import uos as os
                except builtins.ImportError:
                    import os
                import sys

                @builtins.classmethod
                def print_mgmt_value(cls, obj):
                    cls.builtins.print({mgmt_start!r}, cls.builtins.repr(obj), {mgmt_end!r}, sep='', end='')
            """
            ).format(
                mgmt_start=MGMT_VALUE_START.decode(ENCODING),
                mgmt_end=MGMT_VALUE_END.decode(ENCODING),
            )
            + "\n"
        ).lstrip()
        self._execute_without_output(script)

    def fetch_sys_path(self) -> List[str]:
        return self._evaluate("__pipkin_helper.sys.path")

    def fetch_sys_implementation(self) -> Tuple[str, str, int]:
        return self._evaluate("__pipkin_helper.builtins.tuple(__pipkin_helper.sys.implementation)")

    def get_user_packages_path(self) -> Optional[str]:
        return None

    def read_file(self, path: str) -> bytes:
        hex_mode = self._should_hexlify(path)

        open_script = f"__pipkin_fp = __pipkin_helper.builtins.open({path!r}, 'rb')"
        out, err = self._execute_and_capture_output(open_script)

        if (out + err).strip():
            if any(str(nr) in out + err for nr in [errno.ENOENT, errno.ENODEV]):
                raise FileNotFoundError(f"Can't find {path} on target")
            else:
                raise ManagementError(
                    f"Could not open file {path} for reading", script=open_script, out=out, err=err
                )

        if hex_mode:
            self._execute_without_output("from binascii import hexlify as __temp_hexlify")

        block_size = 1024
        num_bytes_read = 0
        blocks = []
        while True:
            if hex_mode:
                block = binascii.unhexlify(
                    self._evaluate("__temp_hexlify(__pipkin_fp.read(%s))" % block_size)
                )
            else:
                block = self._evaluate("__pipkin_fp.read(%s)" % block_size)

            if block:
                blocks.append(block)
                num_bytes_read += len(block)

            if len(block) < block_size:
                break

        self._execute_without_output(
            dedent(
                """
            __pipkin_fp.close()
            del __pipkin_fp
            try:
                del __temp_hexlify
            except:
                pass
            """
            )
        )

        return b"".join(blocks)

    def remove_file_if_exists(self, path: str) -> None:
        self._execute_without_output(
            dedent(
                f"""
            try:
                __pipkin_helper.os.stat({path!r}) and None
            except __pipkin_helper.builtins.OSError:
                pass
            else:
                __pipkin_helper.os.remove({path!r})            
        """
            )
        )

    def remove_dir_if_empty(self, path: str) -> bool:
        return self._evaluate(
            dedent(
                f"""
            if __pipkin_helper.os.listdir({path!r}):
                __pipkin_helper.print_mgmt_value(False)
            else:
                __pipkin_helper.os.remove({path!r})
                __pipkin_helper.print_mgmt_value(True)
        """
            )
        )

    def mkdir_in_existing_parent_exists_ok(self, path: str) -> None:
        self._execute_without_output(
            dedent(
                f"""
            try:
                __pipkin_helper.os.stat({path!r}) and None
            except __pipkin_helper.builtins.OSError:
                __pipkin_helper.os.mkdir({path!r})
        """
            )
        )

    def list_meta_dir_names(self, path: str, dist_name: Optional[str] = None) -> List[str]:
        if dist_name:
            dist_name_condition = f"and name.startswith({dist_name+'-'!r})"
        else:
            dist_name_condition = ""

        return self._evaluate(
            dedent(
                f"""
            try:
                __pipkin_helper.print_mgmt_value([
                    name for name 
                    in __pipkin_helper.os.listdir({path!r}) 
                    if name.endswith('.dist-info') {dist_name_condition}
                ])
            except __pipkin_helper.builtins.OSError as e:
                __pipkin_helper.print_mgmt_value([])
"""
            )
        )

    def _submit_code(self, script: str) -> None:
        assert script

        to_be_sent = script.encode("UTF-8")
        logger.debug("Submitting via %s: %r", self._submit_mode, to_be_sent[:1000])

        # assuming we are already at a prompt, but threads may have produced something extra
        discarded_bytes = self._connection.read_all()
        if discarded_bytes:
            logger.info("Discarding %r", discarded_bytes)

        if self._submit_mode == PASTE_SUBMIT_MODE:
            self._submit_code_via_paste_mode(to_be_sent)
        elif self._submit_mode == RAW_PASTE_SUBMIT_MODE:
            try:
                self._submit_code_via_raw_paste_mode(to_be_sent)
            except RawPasteNotSupportedError:
                logger.warning("Could not use expected raw paste, falling back to paste mode")
                self._submit_mode = PASTE_SUBMIT_MODE
                self._submit_code_via_paste_mode(to_be_sent)
        else:
            self._submit_code_via_raw_mode(to_be_sent)

    def _submit_code_via_paste_mode(self, script_bytes: bytes) -> None:
        # Go to paste mode
        self._ensure_normal_mode()
        self._write(PASTE_MODE_CMD)
        discarded = self._connection.read_until(PASTE_MODE_LINE_PREFIX)
        logger.debug("Discarding %r", discarded)

        # Send script
        while script_bytes:
            block = script_bytes[: self._write_block_size]
            script_bytes = script_bytes[self._write_block_size :]

            # find proper block boundary
            while True:
                expected_echo = block.replace(b"\r\n", b"\r\n" + PASTE_MODE_LINE_PREFIX)
                if (
                    len(expected_echo) > self._write_block_size
                    or block.endswith(b"\r")
                    or len(block) > 2
                    and starts_with_continuation_byte(script_bytes)
                ):
                    # move last byte to the next block
                    script_bytes = block[-1:] + script_bytes
                    block = block[:-1]
                    continue
                else:
                    break

            self._write(block)
            self._connection.read_all_expected(expected_echo, timeout=WAIT_OR_CRASH_TIMEOUT)

        # push and read confirmation
        self._write(EOT)
        expected_confirmation = b"\r\n"
        actual_confirmation = self._connection.read(
            len(expected_confirmation), timeout=WAIT_OR_CRASH_TIMEOUT
        )
        assert actual_confirmation == expected_confirmation, "Expected %r, got %r" % (
            expected_confirmation,
            actual_confirmation,
        )

    def _submit_code_via_raw_mode(self, script_bytes: bytes) -> None:
        self._ensure_raw_mode()

        to_be_written = script_bytes + EOT

        while to_be_written:
            block = to_be_written[self._write_block_size :]
            self._write(block)
            to_be_written = to_be_written[len(block) :]
            if to_be_written:
                time.sleep(self._write_block_delay)

        # fetch command confirmation
        confirmation = self._connection.soft_read(2, timeout=WAIT_OR_CRASH_TIMEOUT)

        if confirmation != OK:
            data = confirmation + self._connection.read_all()
            data += self._connection.read(1, timeout=1, timeout_is_soft=True)
            data += self._connection.read_all()
            logger.error(
                "Could not read command confirmation for script\n\n: %s\n\n" + "Got: %r",
                script_bytes,
                data,
            )
            raise ProtocolError("Could not read command confirmation")

    def _submit_code_via_raw_paste_mode(self, script_bytes: bytes) -> None:
        self._ensure_raw_mode()
        self._connection.set_text_mode(False)
        self._write(RAW_PASTE_COMMAND)
        response = self._connection.soft_read(2, timeout=WAIT_OR_CRASH_TIMEOUT)
        if response != RAW_PASTE_CONFIRMATION:
            # Occasionally, the device initially supports raw paste but later doesn't allow it
            # https://github.com/turcar/turcar/issues/1545
            time.sleep(0.01)
            response += self._connection.read_all()
            if response == FIRST_RAW_PROMPT:
                self._last_prompt = FIRST_RAW_PROMPT
                raise RawPasteNotSupportedError()
            else:
                logger.error("Got %r instead of raw-paste confirmation", response)
                raise ProtocolError("Could not get raw-paste confirmation")

        self._raw_paste_write(script_bytes)
        self._connection.set_text_mode(True)

    def _raw_paste_write(self, command_bytes):
        # Adapted from https://github.com/micropython/micropython/commit/a59282b9bfb6928cd68b696258c0dd2280244eb3#diff-cf10d3c1fe676599a983c0ec85b78c56c9a6f21b2d896c69b3e13f34d454153e

        # Read initial header, with window size.
        data = self._connection.soft_read(2, timeout=2)
        assert len(data) == 2, "Could not read initial header, got %r" % (
            data + self._connection.read_all()
        )
        window_size = data[0] | data[1] << 8
        window_remain = window_size

        # Write out the command_bytes data.
        i = 0
        while i < len(command_bytes):
            while window_remain == 0 or not self._connection.incoming_is_empty():
                data = self._connection.soft_read(1, timeout=WAIT_OR_CRASH_TIMEOUT)
                if data == b"\x01":
                    # Device indicated that a new window of data can be sent.
                    window_remain += window_size
                elif data == b"\x04":
                    # Device indicated abrupt end, most likely a syntax error.
                    # Acknowledge it and finish.
                    self._write(b"\x04")
                    logger.debug(
                        "Abrupt end of raw paste submit after submitting %s bytes out of %s",
                        i,
                        len(command_bytes),
                    )
                    return
                else:
                    # Unexpected data from device.
                    logger.error("Unexpected read during raw paste: %r", data)
                    raise ProtocolError("Unexpected read during raw paste")
            # Send out as much data as possible that fits within the allowed window.
            b = command_bytes[i : min(i + window_remain, len(command_bytes))]
            self._write(b)
            window_remain -= len(b)
            i += len(b)

        # Indicate end of data.
        self._write(b"\x04")

        # Wait for device to acknowledge end of data.
        data = self._connection.soft_read_until(b"\x04", timeout=WAIT_OR_CRASH_TIMEOUT)
        if not data.endswith(b"\x04"):
            logger.error("Could not complete raw paste. Ack: %r", data)
            raise ProtocolError("Could not complete raw paste")

    def _ensure_raw_mode(self):
        if self._last_prompt in [
            RAW_PROMPT,
            EOT + RAW_PROMPT,
            FIRST_RAW_PROMPT,
        ]:
            return
        logger.debug("requesting raw mode at %r", self._last_prompt)

        # assuming we are currently on a normal prompt
        self._write(RAW_MODE_CMD)
        self._log_output_until_active_prompt()
        if self._last_prompt == NORMAL_PROMPT:
            # Don't know why this happens sometimes (e.g. when interrupting a Ctrl+D or restarted
            # program, which is outputting text on ESP32)
            logger.info("Found normal prompt instead of expected raw prompt. Trying again.")
            self._write(RAW_MODE_CMD)
            time.sleep(0.5)
            self._log_output_until_active_prompt()

        if self._last_prompt != FIRST_RAW_PROMPT:
            logger.error(
                "Could not enter raw prompt, got %r",
                self._last_prompt,
            )
            raise ProtocolError("Could not enter raw prompt")

    def _ensure_normal_mode(self, force=False):
        if self._last_prompt == NORMAL_PROMPT and not force:
            return

        logger.debug("requesting normal mode at %r", self._last_prompt)
        self._write(NORMAL_MODE_CMD)
        self._log_output_until_active_prompt()
        assert self._last_prompt == NORMAL_PROMPT, (
            "Could not get normal prompt, got %r" % self._last_prompt
        )

    def _log_output_until_active_prompt(self, timeout: float = WAIT_OR_CRASH_TIMEOUT) -> None:
        def collect_output(data, stream):
            if data:
                logger.info("Discarding %s: %r", stream, data)

        self._process_output_until_active_prompt(collect_output, timeout=timeout)

    def _capture_output_until_active_prompt(self, timeout: float) -> Tuple[str, str]:
        output = {"stdout": "", "stderr": ""}

        def collect_output(data, stream):
            output[stream] += data

        self._process_output_until_active_prompt(collect_output, timeout=timeout)

        return output["stdout"], output["stderr"]

    def _process_output_until_active_prompt(
        self,
        output_consumer: OutputConsumer,
        timeout: float,
    ):
        PROMPT_MARKERS = [NORMAL_PROMPT, EOT + RAW_PROMPT, FIRST_RAW_PROMPT]
        PROMPT_MARKERS_RE = re.compile(
            b"|".join(
                map(
                    re.escape,
                    PROMPT_MARKERS,
                )
            )
        )

        start_time = time.time()

        while True:
            spent_time = time.time() - start_time
            time_left = max(timeout - spent_time, 0.0)
            data = self._connection.read_until(PROMPT_MARKERS_RE, timeout=time_left)
            assert any(data.endswith(marker) for marker in PROMPT_MARKERS)

            for prompt in PROMPT_MARKERS:
                if data.endswith(prompt):
                    self._last_prompt = prompt
                    content = data[: -len(self._last_prompt)]
                    if EOT in content:
                        out, err = content.split(EOT, maxsplit=1)
                    elif TRACEBACK_MARKER in content:
                        out, err = content.split(TRACEBACK_MARKER, maxsplit=1)
                        err = TRACEBACK_MARKER + err
                    else:
                        out = content
                        err = b""
                    output_consumer(out.decode(ENCODING), "stdout")
                    if err:
                        output_consumer(err.decode(ENCODING), "stderr")
                    break

            # Check if it's really active prompt
            follow_up = self._connection.soft_read(1, timeout=0.01)
            if follow_up == ESC:
                # See if it's followed by a OSC code, like the one output by CircuitPython 8
                follow_up += self._connection.soft_read_until(ST)
                if follow_up.endswith(ST):
                    logger.debug("Dropping OSC sequence %r", follow_up)
                follow_up = b""
            if follow_up:
                # Nope, the prompt is not active.
                # (Actually it may be that a background thread has produced this follow up,
                # but this would be too hard to consider.)
                # Don't output yet, because the follow-up may turn into another prompt,
                # and they can be captured all together.
                self._connection.unread(follow_up)
                output_consumer(self._last_prompt.decode(ENCODING), "stdout")
            else:
                break

    def _evaluate(self, script: str) -> Any:
        """Evaluate the output of the script or raise error, if anything looks wrong.

        Adds printing code if the script contains single expression and doesn't
        already contain printing code"""
        try:
            ast.parse(script, mode="eval")
            prefix = "__pipkin_helper.print_mgmt_value("
            suffix = ")"
            if not script.strip().startswith(prefix):
                script = prefix + script + suffix
        except SyntaxError:
            pass

        out, err = self._execute_and_capture_output(script)
        if err:
            raise ManagementError("Script produced errors", script, out, err)
        elif (
            MGMT_VALUE_START.decode(ENCODING) not in out
            or MGMT_VALUE_END.decode(ENCODING) not in out
        ):
            raise ManagementError("Management markers missing", script, out, err)

        start_token_pos = out.index(MGMT_VALUE_START.decode(ENCODING))
        end_token_pos = out.index(MGMT_VALUE_END.decode(ENCODING))

        # a thread or IRQ handler may have written something before or after mgmt value
        prefix = out[:start_token_pos]
        value_str = out[start_token_pos + len(MGMT_VALUE_START) : end_token_pos]
        suffix = out[end_token_pos + len(MGMT_VALUE_END) :]

        try:
            value = ast.literal_eval(value_str)
        except Exception as e:
            raise ManagementError("Could not parse management response", script, out, err) from e

        if prefix:
            logger.warning("Eval output had unexpected prefix: %r", prefix)
        if suffix:
            logger.warning("Eval output had unexpected suffix: %r", suffix)

        return value

    def _write(self, data: bytes) -> None:
        if (
            RAW_MODE_CMD in data
            or NORMAL_MODE_CMD in data
            or INTERRUPT_CMD in data
            or EOT in data
            or PASTE_MODE_CMD in data
        ):
            logger.debug("Sending ctrl chars: %r", data)
        num_bytes = self._connection.write(data)
        assert num_bytes == len(data)

    def _should_hexlify(self, path):
        if "binascii" not in self._builtin_modules and "ubinascii" not in self._builtin_modules:
            return False

        for ext in (".py", ".txt", ".csv", "METADATA", "RECORD"):
            if path.lower().endswith(ext):
                return False

        return True

    def _execute_without_output(self, script: str, timeout: float = WAIT_OR_CRASH_TIMEOUT) -> None:
        """Meant for management tasks."""
        out, err = self._execute_and_capture_output(script, timeout=timeout)
        if out or err:
            raise ManagementError("Command output was not empty", script, out, err)

    def _execute_and_capture_output(
        self, script: str, timeout: float = WAIT_OR_CRASH_TIMEOUT
    ) -> Tuple[str, str]:
        output_lists: Dict[str, List[str]] = {"stdout": [], "stderr": []}

        def consume_output(data, stream_name):
            assert isinstance(data, str)
            output_lists[stream_name].append(data)

        self._execute_with_consumer(script, consume_output, timeout=timeout)
        result = ["".join(output_lists[name]) for name in ["stdout", "stderr"]]
        return result[0], result[1]

    def _execute_with_consumer(
        self, script, output_consumer: OutputConsumer, timeout: float
    ) -> None:
        self._submit_code(script)
        self._process_output_until_active_prompt(output_consumer, timeout=timeout)


class SerialPortAdapter(BareMetalAdapter):
    def __init__(
        self,
        connection: SerialConnection,
        submit_mode: Optional[str] = None,
        write_block_size: Optional[int] = None,
        write_block_delay: Optional[float] = None,
        mount_path: Optional[str] = None,
    ):
        super().__init__(
            connection,
            submit_mode=submit_mode,
            write_block_size=write_block_size,
            write_block_delay=write_block_delay,
        )
        self._mount_path = mount_path
        self._read_only_filesystem = False

    def _internal_path_to_mounted_path(self, target_path: str) -> str:
        assert self._mount_path
        assert target_path.startswith("/")
        return os.path.normpath(os.path.join(self._mount_path, target_path[1:]))

    def write_file_in_existing_dir(self, path: str, content: bytes) -> None:
        start_time = time.time()

        if self._read_only_filesystem:
            self._write_file_via_mount(path, content)

        try:
            self._write_file_via_serial(path, content)
        except ReadOnlyFilesystemError:
            self._read_only_filesystem = True
            self._write_file_via_mount(path, content)

        logger.info("Wrote %s in %.1f seconds", path, time.time() - start_time)

    def _write_file_via_mount(
        self,
        target_path: str,
        content: bytes,
    ) -> None:
        mounted_target_path = self._internal_path_to_mounted_path(target_path)
        with open(mounted_target_path, "wb") as f:
            bytes_written = 0
            block_size = 4 * 1024
            to_be_written = content
            while to_be_written:
                block = to_be_written[:block_size]
                bytes_written += f.write(block)
                assert bytes_written
                f.flush()
                os.fsync(f)
                to_be_written = to_be_written[block_size:]

        assert bytes_written == len(content)

    def _write_file_via_serial(
        self, target_path: str, content: bytes, can_hexlify: bool = True
    ) -> None:
        out, err = self._execute_and_capture_output(
            dedent(
                """
                __pipkin_path = '{path}'
                __pipkin_written = 0
                __pipkin_fp = __pipkin_helper.builtins.open(__pipkin_path, 'wb')
            """
            ).format(path=target_path),
        )

        if self._contains_read_only_error(out + err):
            raise ReadOnlyFilesystemError()
        elif out + err:
            raise OSError(
                "Could not open file %s for writing, output:\n%s" % (target_path, out + err)
            )

        # Define function to allow shorter write commands
        hex_mode = self._should_hexlify(target_path)
        if hex_mode:
            self._execute_without_output(
                dedent(
                    """
                from binascii import unhexlify as __pipkin_unhex
                def __W(x):
                    global __pipkin_written
                    __pipkin_written += __pipkin_fp.write(__pipkin_unhex(x))
                    __pipkin_fp.flush()
                    if __pipkin_helper.builtins.hasattr(__pipkin_helper.os, "sync"):
                        __pipkin_helper.os.sync()
            """
                )
            )
        else:
            self._execute_without_output(
                dedent(
                    """
                def __W(x):
                    global __pipkin_written
                    __pipkin_written += __pipkin_fp.write(x)
                    __pipkin_fp.flush()
                    if __pipkin_helper.builtins.hasattr(__pipkin_helper.os, "sync"):
                        __pipkin_helper.os.sync()
            """
                )
            )

        bytes_sent = 0
        block_size = 1024

        to_be_written = content
        while to_be_written:
            block = to_be_written[:block_size]
            if hex_mode:
                script = "__W(%r)" % binascii.hexlify(block)
            else:
                script = "__W(%r)" % block
            out, err = self._execute_and_capture_output(script)
            if out or err:
                logger.error("Writing file produced unexpected output (%r) or error (%r)", out, err)
                raise UserError(
                    "Could not write next block after having written %d bytes to %s"
                    % (bytes_sent, target_path)
                )

            bytes_sent += len(block)
            to_be_written = to_be_written[block_size:]

        bytes_received = self._evaluate("__pipkin_written")

        if bytes_received != bytes_sent:
            raise OSError("Expected %d written bytes but wrote %d" % (bytes_sent, bytes_received))

        # clean up
        self._execute_without_output(
            dedent(
                """
                try:
                    del __W
                    del __pipkin_written
                    del __pipkin_path
                    __pipkin_fp.close()
                    del __pipkin_fp
                    del __pipkin_result
                    del __pipkin_unhex
                except:
                    pass
            """
            )
        )

    def remove_file_if_exists(self, path: str) -> None:
        if self._read_only_filesystem:
            self._remove_file_via_mount(path)
            return

        try:
            super().remove_file_if_exists(path)
        except ManagementError as e:
            if self._contains_read_only_error(e.out + e.err):
                self._read_only_filesystem = True
                self._remove_file_via_mount(path)
            else:
                raise

    def _remove_file_via_mount(self, target_path: str) -> None:
        logger.info("Removing %s via mount", target_path)
        mounted_target_path = self._internal_path_to_mounted_path(target_path)
        assert os.path.isfile(mounted_target_path)
        os.remove(mounted_target_path)

    def _contains_read_only_error(self, s: str) -> bool:
        canonic_out = s.replace("-", "").lower()
        return (
            "readonly" in canonic_out
            or f"errno {errno.EROFS}" in canonic_out
            or f"oserror: {errno.EROFS}" in canonic_out
        )

    def mkdir_in_existing_parent_exists_ok(self, path: str) -> None:
        if self._read_only_filesystem:
            self._mkdir_via_mount(path)
            return

        try:
            super().mkdir_in_existing_parent_exists_ok(path)
        except ManagementError as e:
            if self._contains_read_only_error(e.out + e.err):
                self._read_only_filesystem = True
                self._mkdir_via_mount(path)
            else:
                raise

    def _mkdir_via_mount(self, path: str) -> bool:
        mounted_path = self._internal_path_to_mounted_path(path)
        if not os.path.isdir(mounted_path):
            assert not os.path.exists(mounted_path)
            os.mkdir(mounted_path, 0o755)
            return True
        else:
            return False

    def remove_dir_if_empty(self, path: str) -> bool:
        if self._read_only_filesystem:
            return self._remove_dir_if_empty_via_mount(path)

        try:
            return super().remove_dir_if_empty(path)
        except ManagementError as e:
            if self._contains_read_only_error(e.out + e.err):
                self._read_only_filesystem = True
                return self._remove_dir_if_empty_via_mount(path)
            else:
                raise

    def _remove_dir_if_empty_via_mount(self, path: str) -> bool:
        mounted_path = self._internal_path_to_mounted_path(path)
        if os.listdir(mounted_path):
            return False
        else:
            os.rmdir(mounted_path)
            return True


class WebReplAdapter(BareMetalAdapter):
    def write_file_in_existing_dir(self, path: str, content: bytes) -> None:
        """
        Adapted from https://github.com/micropython/webrepl/blob/master/webrepl_cli.py
        """
        dest_fname = path.encode("utf-8")
        rec = struct.pack(
            WEBREPL_REQ_S, b"WA", WEBREPL_PUT_FILE, 0, 0, len(content), len(dest_fname), dest_fname
        )
        self._connection.set_text_mode(False)
        try:
            self._write(rec[:10])
            self._write(rec[10:])
            assert self._read_websocket_response() == 0

            to_be_written = content
            block_size = 1024
            while to_be_written:
                block = to_be_written[:block_size]
                self._write(block)
                to_be_written = to_be_written[block_size:]

            assert self._read_websocket_response() == 0
        finally:
            self._connection.set_text_mode(True)

    def _read_websocket_response(self):
        data = self._connection.read(4)
        sig, code = struct.unpack("<2sH", data)
        assert sig == b"WB"
        return code


class RawPasteNotSupportedError(RuntimeError):
    pass


class ReadOnlyFilesystemError(OSError):
    pass



import os
import re
import sys
import subprocess
import requests
import tkinter as tk
from tkinter import messagebox


class CheackUpdates:
    def __init__(self):
        self.os_type = sys.platform
        self.root = tk.Tk()
        self.url = f"https://pypi.org/pypi/turcar/json"

    def is_windows(self):
        return self.os_type.startswith('win')

    def is_linux(self):
        return self.os_type.startswith('linux')

    def is_mac(self):
        return self.os_type.startswith('darwin')

    # 检查版本
    def cheack_version(self):
        response = requests.get(self.url)
        if response.status_code == 200:
            data = response.json()
            # 获取当前脚本所在目录
            version = data["info"]["version"]
            # 获取当前脚本所在目录
            current_dir = os.path.dirname(os.path.abspath(__file__))
            # 指定要获取内容的文件名
            file_name = "VERSION"
            # 拼接文件路径
            file_path = os.path.join(current_dir, file_name)
            # 空校验
            if os.path.exists(file_path):
                # 打开文件并读取内容
                with open(file_path, 'r') as file:
                    content = file.read()
                    # print(f"File: {file_name}")
                    print(f"VERSION: {content}")
                    if version != content:
                        # 创建一个确认更新的弹窗
                        root = self.root
                        root.withdraw()  # 隐藏主窗口
                        result = messagebox.askokcancel("确认更新", "发现新版本，是否更新？")
                        if result:
                            # 处理更新脚本中的版本号
                            update_bash("turcar-update.bash", version)
                            # 执行更新脚本
                            # bash_command = '/turcar-update.bash'
                            # execute_bash_command(bash_command)
                            # 更新成功后本地版本号更新
                            with open(file_name, 'w') as wr_file:
                                wr_file.write(version)
                            print("更新成功")
                        else:
                            print("取消更新")

            return version
        else:
            return None


# 更新脚本中的版本号
def update_bash(filename, new_version):
    """
    更新脚本中的版本号

    Args:
        filename: 要更新的脚本文件名
        new_version: 新的版本号，格式为 x.x.x
    """
    # 使用正则表达式匹配并替换版本号
    with open(filename, 'r') as file:
        content = file.read()
        new_content = re.sub(r'(VERSION=)\d+\.\d+\.\d+', r'\g<1>' + new_version, content)

    # 将更新后的内容写回脚本文件
    with open(filename, 'w') as file:
        file.write(new_content)


# 执行 Bash 命令
def execute_bash_command(command):
    """
    执行 Bash 命令

    Args:
        command: 要执行的 Bash 命令
    """
    subprocess.run(command, shell=True)



import logging
import subprocess
import sys
import traceback
from typing import List, Optional

from pipkin.adapters import Adapter, DummyAdapter, create_adapter
from pipkin.common import ManagementError, UserError
from pipkin.session import Session

logger = logging.getLogger("pipkin")

__version__ = "2.0b2"


def error(msg):
    msg = "ERROR: " + msg
    print(msg, file=sys.stderr)

    return 1


def main(raw_args: Optional[List[str]] = None) -> int:
    from pipkin import parser

    args = parser.parse_arguments(raw_args)

    if args.verbose:
        logging_level = logging.DEBUG
    elif args.quiet:
        logging_level = logging.ERROR
    else:
        logging_level = logging.INFO

    logger.setLevel(logging_level)
    logger.propagate = True
    console_handler = logging.StreamHandler(sys.stderr)
    console_handler.setLevel(logging_level)
    logger.addHandler(console_handler)

    args_dict = vars(args)

    try:
        adapter: Adapter
        if args.command == "cache":
            adapter = DummyAdapter()
        else:
            adapter = create_adapter(**args_dict)
        session = Session(adapter)
        try:
            command_handler = getattr(session, args.command)
            command_handler(**args_dict)
        finally:
            session.close()
    except KeyboardInterrupt:
        return 1
    except ManagementError as e:
        logger.error(traceback.format_exc())
        logger.error("SCRIPT: %r", e.script)
        logger.error("OUT=%r", e.out)
        logger.error("ERR=%r", e.err)
    except UserError as e:
        return error(str(e))
    except subprocess.CalledProcessError:
        # assuming the subprocess (pip) already printed the error
        return 1

    return 0



"""
Module: '_api.motorpair' on micropython-v1.14-893-lego learning system hub
"""
# MCU: {'machine': 'LEGO Technic Large Hub with STM32F413xx', 'sysname': 'LEGO Technic Large Hub', 'platform': 'LEGO Learning System Hub', 'nodename': 'LEGO Learning System Hub', 'ver': 'v1.14-893', 'release': '1.14.0', 'name': 'micropython', 'family': 'micropython', 'port': 'LEGO Learning System Hub', 'version': '1.14.0', 'mpy': 517, 'build': '893'}
# Stubber: 1.7.2
from typing import Any


class MotorPair():
    def __init__(self, *argv, **kwargs) -> None:
        """"""

    BRAKE = 'brake' # type: str
    CM = 'cm' # type: str
    COAST = 'coast' # type: str
    DEGREES = 'degrees' # type: str
    HOLD = 'hold' # type: str
    IN = 'in' # type: str
    ROTATIONS = 'rotations' # type: str
    SECONDS = 'seconds' # type: str
    def get_default_speed(self, *args, **kwargs) -> Any:
        """"""

    def move(self, *args, **kwargs) -> Any:
        """"""

    def move_tank(self, *args, **kwargs) -> Any:
        """"""

    def set_default_speed(self, *args, **kwargs) -> Any:
        """"""

    def set_motor_rotation(self, *args, **kwargs) -> Any:
        """"""

    def set_stop_action(self, *args, **kwargs) -> Any:
        """"""

    def start(self, *args, **kwargs) -> Any:
        """"""

    def start_at_power(self, *args, **kwargs) -> Any:
        """"""

    def start_tank(self, *args, **kwargs) -> Any:
        """"""

    def start_tank_at_power(self, *args, **kwargs) -> Any:
        """"""

    def stop(self, *args, **kwargs) -> Any:
        """"""

    def was_interrupted(self, *args, **kwargs) -> Any:
        """"""

PORTS = {} # type: dict
def clamp_power(*args, **kwargs) -> Any:
    """"""

def clamp_speed(*args, **kwargs) -> Any:
    """"""

def clamp_steering(*args, **kwargs) -> Any:
    """"""

def from_steering(*args, **kwargs) -> Any:
    """"""

system : Any ## <class 'System'> = <System object at 20028790>
def wait_for_async(*args, **kwargs) -> Any:
    """"""

